#!/usr/bin/python

# Copyright (c) 2014, Steve Smith <ssmith@atlassian.com>
# Atlassian open-source approval reference OSR-76.
#
# Copyright (c) 2020, Per Abildgaard Toft <per@minfejl.dk> Search and update function
# Copyright (c) 2021, Brandon McNama <brandonmcnama@outlook.com> Issue attachment functionality
# Copyright (c) 2022, Hugo Prudente <hugo.kenshin+oss@gmail.com> Worklog functionality
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: jira
short_description: Create and modify issues in a JIRA instance
description:
  - Create and modify issues in a JIRA instance.
extends_documentation_fragment:
  - community.general._attributes

attributes:
  check_mode:
    support: none
  diff_mode:
    support: none

options:
  uri:
    type: str
    required: true
    description:
      - Base URI for the JIRA instance.
  operation:
    type: str
    required: true
    aliases: [command]
    choices: [attach, comment, create, edit, fetch, link, search, transition, update, worklog]
    description:
      - The operation to perform.
      - V(worklog) was added in community.general 6.5.0.
  username:
    type: str
    description:
      - The username to log-in with.
      - Must be used with O(password). Mutually exclusive with O(token).
  password:
    type: str
    description:
      - The password to log-in with.
      - Must be used with O(username). Mutually exclusive with O(token).
  token:
    type: str
    description:
      - The personal access token to log-in with.
      - Mutually exclusive with O(username) and O(password).
    version_added: 4.2.0
  client_cert:
    type: path
    description:
      - Client certificate if required.
      - In addition to O(username) and O(password) or O(token). Not mutually exclusive.
    version_added: 10.4.0
  client_key:
    type: path
    description:
      - Client certificate key if required.
      - In addition to O(username) and O(password) or O(token). Not mutually exclusive.
    version_added: 10.4.0

  project:
    type: str
    description:
      - The project for this operation. Required for issue creation.
  summary:
    type: str
    description:
      - The issue summary, where appropriate.
      - Note that JIRA may not allow changing field values on specific transitions or states.
  description:
    type: str
    description:
      - The issue description, where appropriate.
      - Note that JIRA may not allow changing field values on specific transitions or states.
  issuetype:
    type: str
    description:
      - The issue type, for issue creation.
  issue:
    type: str
    description:
      - An existing issue key to operate on.
    aliases: ['ticket']

  comment:
    type: str
    description:
      - The comment text to add.
      - Note that JIRA may not allow changing field values on specific transitions or states.
  comment_visibility:
    type: dict
    description:
      - Used to specify comment comment visibility.
      - See
        U(https://developer.atlassian.com/cloud/jira/platform/rest/v2/api-group-issue-comments/#api-rest-api-2-issue-issueidorkey-comment-post)
        for details.
    suboptions:
      type:
        description:
          - Use O(comment_visibility.type) to specify which of the JIRA visibility restriction types is used.
        type: str
        required: true
        choices: [group, role]
      value:
        description:
          - Specify value corresponding to the type of visibility restriction. For example name of the group or role.
        type: str
        required: true
    version_added: '3.2.0'

  status:
    type: str
    description:
      - Only used when O(operation) is V(transition), and a bit of a misnomer, it actually refers to the transition name.
      - This is mutually exclusive with O(status_id).
  status_id:
    type: str
    description:
      - Only used when O(operation) is V(transition), and refers to the transition ID.
      - This is mutually exclusive with O(status).
    version_added: 10.3.0
  assignee:
    type: str
    description:
      - Sets the assignee when O(operation) is V(create), V(transition), or V(edit).
      - Jira Cloud requires account IDs for all user-type fields. When O(cloud=true), a string value containing C(@) is
        resolved as an email to an account ID automatically. A string without C(@) is assumed to be an account ID.
        See also O(account_id) for a dedicated assignee account ID parameter.
        Added in community.general 12.6.0.
      - Jira Server and Jira Data Center may still accept O(assignee) as a username.
      - Note that JIRA may not allow changing field values on specific transitions or states.
  account_id:
    type: str
    description:
      - Sets the account identifier for the assignee when O(operation) is V(create), V(transition), or V(edit).
      - Note that JIRA may not allow changing field values on specific transitions or states.
    version_added: 2.5.0

  linktype:
    type: str
    description:
      - Set type of link, when action 'link' selected.
  inwardissue:
    type: str
    description:
      - Set issue from which link is created.
  outwardissue:
    type: str
    description:
      - Set issue to which link is created.
  fields:
    type: dict
    description:
      - This is a free-form data structure that can contain arbitrary data. This is passed directly to the JIRA REST API (possibly
        after merging with other required data, as when passed to create). See examples for more information, and the JIRA
        REST API for the structure required for various fields.
      - When O(cloud=true), user-type fields (V(assignee), V(reporter), and any fields listed in O(custom_user_fields)) that
        contain a string value with C(@) are automatically resolved from email to account ID. String values without C(@) are
        assumed to be account IDs. Added in community.general 12.6.0.
      - When passed to comment, the data structure is merged at the first level since community.general 4.6.0. Useful to add
        JIRA properties for example.
      - Note that JIRA may not allow changing field values on specific transitions or states.
    default: {}
  jql:
    description:
      - Query JIRA in JQL Syntax, for example V("CMDB Hostname" = test.example.com).
    type: str
    version_added: '0.2.0'

  maxresults:
    description:
      - Limit the result of O(operation=search). If no value is specified, the default JIRA limit is used.
      - Used when O(operation=search) only, ignored otherwise.
    type: int
    version_added: '0.2.0'

  timeout:
    type: float
    description:
      - Set timeout, in seconds, on requests to JIRA API.
    default: 10

  validate_certs:
    description:
      - Require valid SSL certificates (set to V(false) if you would like to use self-signed certificates).
    default: true
    type: bool

  cloud:
    description:
      - Enable when using Jira Cloud.
      - When set to V(true), the module uses Jira Cloud compatible behavior.
      - When set to V(false) (the default), the module uses Jira Server / Data Center compatible behavior.
      - In the future, this option might affect additional operations for Jira Cloud compatibility.
    type: bool
    default: false
    version_added: '12.6.0'

  custom_user_fields:
    description:
      - A list of field names (for example V(customfield_10050)) that hold user values.
      - When O(cloud=true) and a listed field is present in O(fields) as a string containing C(@), the module resolves the
        email to a Jira Cloud account ID automatically.
      - The built-in user fields O(assignee) and V(reporter) are always resolved when present; list here any further
        user-typed fields (from other Jira products or extensions) that should receive the same resolution.
    type: list
    elements: str
    default: []
    version_added: '12.6.0'

  attachment:
    type: dict
    version_added: 2.5.0
    description:
      - Information about the attachment being uploaded.
    suboptions:
      filename:
        required: true
        type: path
        description:
          - The path to the file to upload (from the remote node) or, if O(attachment.content) is specified, the filename
            to use for the attachment.
      content:
        type: str
        description:
          - The Base64 encoded contents of the file to attach. If not specified, the contents of O(attachment.filename) is
            used instead.
      mimetype:
        type: str
        description:
          - The MIME type to supply for the upload. If not specified, best-effort detection is performed.
notes:
  - Currently this only works with basic-auth, or tokens.
  - To use with JIRA Cloud, pass the login e-mail as the O(username) and the API token as O(password).
  - Jira Cloud has removed the legacy C(/rest/api/2/search) endpoint. Set O(cloud=true) when using
    O(operation=search) against Jira Cloud so that the module uses the replacement C(/rest/api/2/search/jql) endpoint instead.
author:
  - "Steve Smith (@tarka)"
  - "Per Abildgaard Toft (@pertoft)"
  - "Brandon McNama (@DWSR)"
"""

EXAMPLES = r"""
# Create a new issue and add a comment to it:
- name: Create an issue
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    project: ANS
    operation: create
    summary: Example Issue
    description: Created using Ansible
    issuetype: Task
  args:
    fields:
      customfield_13225: "test"
      customfield_12931: {"value": "Test"}
  register: issue

- name: Comment on issue
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key }}'
    operation: comment
    comment: A comment added by Ansible

- name: Comment on issue with restricted visibility
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key }}'
    operation: comment
    comment: A comment added by Ansible
    comment_visibility:
      type: role
      value: Developers

- name: Comment on issue with property to mark it internal
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key }}'
    operation: comment
    comment: A comment added by Ansible
    fields:
      properties:
        - key: 'sd.public.comment'
          value:
            internal: true

# Add an workog to an existing issue
- name: Worklog on issue
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key }}'
    operation: worklog
    comment: A worklog added by Ansible
    fields:
      timeSpentSeconds: 12000

- name: Workflow on issue with comment restricted visibility
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key }}'
    operation: worklog
    comment: A worklog added by Ansible
    comment_visibility:
      type: role
      value: Developers
    fields:
      timeSpentSeconds: 12000

- name: Workflow on issue with comment property to mark it internal
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key }}'
    operation: worklog
    comment: A worklog added by Ansible
    fields:
      properties:
        - key: 'sd.public.comment'
          value:
            internal: true
      timeSpentSeconds: 12000

# Assign an existing issue using edit
- name: Assign an issue using free-form fields
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key}}'
    operation: edit
    assignee: ssmith

# Create an issue with an existing assignee
- name: Create an assigned issue
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    project: ANS
    operation: create
    summary: Assigned issue
    description: Created and assigned using Ansible
    issuetype: Task
    assignee: ssmith

# Assign an issue on Jira Cloud using an email address
# (the module resolves the email to an account ID automatically)
- name: Assign an issue on Jira Cloud using email
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key }}'
    operation: edit
    cloud: true
    assignee: user@example.com

# Create an issue on Jira Cloud with reporter and a custom user field
# resolved from email addresses to account IDs
- name: Create an issue on Jira Cloud with user field resolution
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    project: ANS
    operation: create
    summary: Example Issue
    description: Created using Ansible
    issuetype: Task
    cloud: true
    assignee: assignee@example.com
    custom_user_fields:
      - customfield_10050
  args:
    fields:
      reporter: reporter@example.com
      customfield_10050: approver@example.com

# Edit an issue
- name: Set the labels on an issue using free-form fields
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key }}'
    operation: edit
  args:
    fields:
      labels:
        - autocreated
        - ansible

# Updating a field using operations: add, set & remove
- name: Change the value of a Select dropdown
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key }}'
    operation: update
  args:
    fields:
      customfield_12931: ['set': {'value': 'Virtual'}]
      customfield_13820: ['set': {'value': 'Manually'}]
  register: cmdb_issue
  delegate_to: localhost


# Retrieve metadata for an issue and use it to create an account
- name: Get an issue
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    project: ANS
    operation: fetch
    issue: ANS-63
  register: issue

# Search for an issue on Jira Data Center / Server
# You can limit the search for specific fields by adding optional args. Note! It must be a dict, hence, lastViewed: null
- name: Search for an issue
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    project: ANS
    operation: search
    maxresults: 10
    jql: project=cmdb AND cf[13225]="test"
  args:
    fields:
      lastViewed:
  register: issue

# Search for an issue on Jira Cloud (uses the /search/jql endpoint)
- name: Search for an issue on Jira Cloud
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    project: ANS
    operation: search
    cloud: true
    maxresults: 10
    jql: project=cmdb AND cf[13225]="test"
  args:
    fields:
      lastViewed:
  register: issue

- name: Create a unix account for the reporter
  become: true
  user:
    name: '{{ issue.meta.fields.creator.name }}'
    comment: '{{ issue.meta.fields.creator.displayName }}'

# You can get list of valid linktypes at /rest/api/2/issueLinkType
# url of your jira installation.
- name: Create link from HSP-1 to MKY-1
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    operation: link
    linktype: Relates
    inwardissue: HSP-1
    outwardissue: MKY-1

# Transition an issue
- name: Resolve the issue
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: '{{ issue.meta.key }}'
    operation: transition
    status: Resolve Issue
    account_id: 112233445566778899aabbcc
    fields:
      resolution:
        name: Done
      description: I am done! This is the last description I will ever give you.

# Attach a file to an issue
- name: Attach a file
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    issue: HSP-1
    operation: attach
    attachment:
      filename: topsecretreport.xlsx

# Use username, password and client certificate authentification
- name: Create an issue
  community.general.jira:
    uri: '{{ server }}'
    username: '{{ user }}'
    password: '{{ pass }}'
    client_cert: '{{ path/to/client-cert }}'
    client_key: '{{ path/to/client-key }}'

# Use token and client certificate authentification
- name: Create an issue
  community.general.jira:
    uri: '{{ server }}'
    token: '{{ token }}'
    client_cert: '{{ path/to/client-cert }}'
    client_key: '{{ path/to/client-key }}'
"""

import base64
import binascii
import json
import mimetypes
import os
import random
import string
import traceback
from urllib.parse import quote
from urllib.request import pathname2url

from ansible.module_utils.common.text.converters import to_bytes, to_native, to_text
from ansible.module_utils.urls import fetch_url

from ansible_collections.community.general.plugins.module_utils._module_helper import StateModuleHelper, cause_changes


class JIRA(StateModuleHelper):
    USER_FIELDS = ["assignee", "reporter"]

    module = dict(
        argument_spec=dict(
            attachment=dict(
                type="dict",
                options=dict(
                    content=dict(type="str"), filename=dict(type="path", required=True), mimetype=dict(type="str")
                ),
            ),
            uri=dict(type="str", required=True),
            operation=dict(
                type="str",
                choices=[
                    "attach",
                    "create",
                    "comment",
                    "edit",
                    "update",
                    "fetch",
                    "transition",
                    "link",
                    "search",
                    "worklog",
                ],
                aliases=["command"],
                required=True,
            ),
            username=dict(type="str"),
            password=dict(type="str", no_log=True),
            token=dict(type="str", no_log=True),
            client_cert=dict(type="path"),
            client_key=dict(type="path"),
            project=dict(
                type="str",
            ),
            summary=dict(
                type="str",
            ),
            description=dict(
                type="str",
            ),
            issuetype=dict(
                type="str",
            ),
            issue=dict(type="str", aliases=["ticket"]),
            comment=dict(
                type="str",
            ),
            comment_visibility=dict(
                type="dict",
                options=dict(
                    type=dict(type="str", choices=["group", "role"], required=True),
                    value=dict(type="str", required=True),
                ),
            ),
            status=dict(
                type="str",
            ),
            status_id=dict(
                type="str",
            ),
            assignee=dict(
                type="str",
            ),
            fields=dict(default={}, type="dict"),
            linktype=dict(
                type="str",
            ),
            inwardissue=dict(
                type="str",
            ),
            outwardissue=dict(
                type="str",
            ),
            jql=dict(
                type="str",
            ),
            cloud=dict(type="bool", default=False),
            custom_user_fields=dict(type="list", elements="str", default=[]),
            maxresults=dict(type="int"),
            timeout=dict(type="float", default=10),
            validate_certs=dict(default=True, type="bool"),
            account_id=dict(type="str"),
        ),
        mutually_exclusive=[
            ["username", "token"],
            ["password", "token"],
            ["assignee", "account_id"],
            ["status", "status_id"],
        ],
        required_together=[["username", "password"], ["client_cert", "client_key"]],
        required_one_of=[
            ["username", "token"],
        ],
        required_if=(
            ("operation", "attach", ["issue", "attachment"]),
            ("operation", "create", ["project", "issuetype", "summary"]),
            ("operation", "comment", ["issue", "comment"]),
            ("operation", "workflow", ["issue", "comment"]),
            ("operation", "fetch", ["issue"]),
            ("operation", "transition", ["issue"]),
            ("operation", "transition", ["status", "status_id"], True),
            ("operation", "link", ["linktype", "inwardissue", "outwardissue"]),
            ("operation", "search", ["jql"]),
        ),
        supports_check_mode=False,
    )
    state_param = "operation"

    def __init_module__(self):
        self.vars.uri = self.vars.uri.strip("/")
        self.vars.set("restbase", f"{self.vars.uri}/rest/api/2")
        if self.vars.fields is None:
            self.vars.fields = {}
        if self.vars.assignee:
            if self.vars.cloud:
                self.vars.fields["assignee"] = self.vars.assignee
            else:
                self.vars.fields["assignee"] = {"name": self.vars.assignee}
        if self.vars.account_id:
            self.vars.fields["assignee"] = {"accountId": self.vars.account_id}
        self.resolve_user_fields()

    def resolve_user_fields(self):
        if not self.vars.cloud or not self.vars.fields:
            return
        user_fields = self.USER_FIELDS + self.vars.custom_user_fields
        for field_name in user_fields:
            value = self.vars.fields.get(field_name)
            if not isinstance(value, str):
                continue
            if "@" in value:
                account_id = self.resolve_account_id(value, field_name)
                self.vars.fields[field_name] = {"accountId": account_id}
            else:
                self.vars.fields[field_name] = {"accountId": value}

    def resolve_account_id(self, email, field_name="assignee"):
        url = f"{self.vars.restbase}/user/search?query={quote(email, safe='')}"
        result = self.get(url)
        if not isinstance(result, list) or len(result) != 1:
            count = len(result) if isinstance(result, list) else 0
            self.module.fail_json(
                msg=(
                    f'Failed to resolve field "{field_name}" email "{email}" to a unique '
                    f'Jira Cloud account ID: found "{count}" result(s). '
                    "Specify the account ID directly."
                )
            )
        user = result[0]
        user_email = user.get("emailAddress")
        if (user_email or "").lower() != email.lower():
            self.module.fail_json(
                msg=(
                    f'Failed to resolve field "{field_name}" email "{email}": '
                    f'the email address on the matched account "{user_email}" does not match. '
                    "Specify the account ID directly."
                )
            )
        return user["accountId"]

    @cause_changes(when="success")
    def operation_create(self):
        createfields = {
            "project": {"key": self.vars.project},
            "summary": self.vars.summary,
            "issuetype": {"name": self.vars.issuetype},
        }

        if self.vars.description:
            createfields["description"] = self.vars.description

        # Merge in any additional or overridden fields
        if self.vars.fields:
            createfields.update(self.vars.fields)

        data = {"fields": createfields}
        url = f"{self.vars.restbase}/issue/"
        self.vars.meta = self.post(url, data)

    @cause_changes(when="success")
    def operation_comment(self):
        data = {"body": self.vars.comment}
        # if comment_visibility is specified restrict visibility
        if self.vars.comment_visibility is not None:
            data["visibility"] = self.vars.comment_visibility

        # Use 'fields' to merge in any additional data
        if self.vars.fields:
            data.update(self.vars.fields)

        url = f"{self.vars.restbase}/issue/{self.vars.issue}/comment"
        self.vars.meta = self.post(url, data)

    @cause_changes(when="success")
    def operation_worklog(self):
        data = {"comment": self.vars.comment}
        # if comment_visibility is specified restrict visibility
        if self.vars.comment_visibility is not None:
            data["visibility"] = self.vars.comment_visibility

        # Use 'fields' to merge in any additional data
        if self.vars.fields:
            data.update(self.vars.fields)

        url = f"{self.vars.restbase}/issue/{self.vars.issue}/worklog"
        self.vars.meta = self.post(url, data)

    @cause_changes(when="success")
    def operation_edit(self):
        data = {"fields": self.vars.fields}
        url = f"{self.vars.restbase}/issue/{self.vars.issue}"
        self.vars.meta = self.put(url, data)

    @cause_changes(when="success")
    def operation_update(self):
        data = {
            "update": self.vars.fields,
        }
        url = f"{self.vars.restbase}/issue/{self.vars.issue}"
        self.vars.meta = self.put(url, data)

    def operation_fetch(self):
        url = f"{self.vars.restbase}/issue/{self.vars.issue}"
        self.vars.meta = self.get(url)

    def operation_search(self):
        if self.vars.cloud:
            url = f"{self.vars.restbase}/search/jql?jql={pathname2url(self.vars.jql)}"
        else:
            url = f"{self.vars.restbase}/search?jql={pathname2url(self.vars.jql)}"

        if self.vars.fields:
            fields = self.vars.fields.keys()
            url = f"{url}&fields={'&fields='.join([pathname2url(f) for f in fields])}"
        elif self.vars.cloud:
            url = f"{url}&fields=*navigable"

        if self.vars.maxresults:
            url = f"{url}&maxResults={self.vars.maxresults}"

        self.vars.meta = self.get(url)

    @cause_changes(when="success")
    def operation_transition(self):
        # Find the transition id
        turl = f"{self.vars.restbase}/issue/{self.vars.issue}/transitions"
        tmeta = self.get(turl)

        tid = None
        target = None

        if self.vars.status is not None:
            target = self.vars.status.strip()
        elif self.vars.status_id is not None:
            tid = self.vars.status_id.strip()

        for t in tmeta["transitions"]:
            if target is not None:
                if t["name"] == target:
                    tid = t["id"]
                    break
            else:
                if tid == t["id"]:
                    break
        else:
            if target is not None:
                raise ValueError(f"Failed find valid transition for '{target}'")
            else:
                raise ValueError(f"Failed find valid transition for ID '{tid}'")

        fields = dict(self.vars.fields)
        if self.vars.summary is not None:
            fields.update({"summary": self.vars.summary})
        if self.vars.description is not None:
            fields.update({"description": self.vars.description})

        # Perform it
        data = {"transition": {"id": tid}, "fields": fields}
        if self.vars.comment is not None:
            data.update(
                {
                    "update": {
                        "comment": [{"add": {"body": self.vars.comment}}],
                    }
                }
            )
        url = f"{self.vars.restbase}/issue/{self.vars.issue}/transitions"
        self.vars.meta = self.post(url, data)

    @cause_changes(when="success")
    def operation_link(self):
        data = {
            "type": {"name": self.vars.linktype},
            "inwardIssue": {"key": self.vars.inwardissue},
            "outwardIssue": {"key": self.vars.outwardissue},
        }
        url = f"{self.vars.restbase}/issueLink/"
        self.vars.meta = self.post(url, data)

    @cause_changes(when="success")
    def operation_attach(self):
        v = self.vars
        filename = v.attachment.get("filename")
        content = v.attachment.get("content")

        if not any((filename, content)):
            raise ValueError("at least one of filename or content must be provided")
        mime = v.attachment.get("mimetype")

        if not os.path.isfile(filename):
            raise ValueError(f"The provided filename does not exist: {filename}")

        content_type, data = self._prepare_attachment(filename, content, mime)

        url = f"{v.restbase}/issue/{v.issue}/attachments"
        return True, self.post(
            url, data, content_type=content_type, additional_headers={"X-Atlassian-Token": "no-check"}
        )

    # Ideally we'd just use prepare_multipart from ansible.module_utils.urls, but
    # unfortunately it does not support specifying the encoding and also defaults to
    # base64. Jira doesn't support base64 encoded attachments (and is therefore not
    # spec compliant. Go figure). I originally wrote this function as an almost
    # exact copypasta of prepare_multipart, but ran into some encoding issues when
    # using the noop encoder. Hand rolling the entire message body seemed to work
    # out much better.
    #
    # https://community.atlassian.com/t5/Jira-questions/Jira-dosen-t-decode-base64-attachment-request-REST-API/qaq-p/916427
    #
    # content is expected to be a base64 encoded string since Ansible doesn't
    # support passing raw bytes objects.
    @staticmethod
    def _prepare_attachment(filename, content=None, mime_type=None):
        def escape_quotes(s):
            return s.replace('"', '\\"')

        boundary = "".join(random.choice(string.digits + string.ascii_letters) for dummy in range(30))
        name = os.path.basename(filename)

        if not mime_type:
            try:
                mime_type = mimetypes.guess_type(filename or "", strict=False)[0] or "application/octet-stream"
            except Exception:
                mime_type = "application/octet-stream"
        main_type, sep, sub_type = mime_type.partition("/")

        if not content and filename:
            with open(to_bytes(filename, errors="surrogate_or_strict"), "rb") as f:
                content = f.read()
        else:
            try:
                content = base64.b64decode(content)
            except binascii.Error as e:
                raise Exception(f"Unable to base64 decode file content: {e}") from e

        lines = [
            f"--{boundary}",
            f'Content-Disposition: form-data; name="file"; filename={escape_quotes(name)}',
            f"Content-Type: {main_type}/{sub_type}",
            "",
            to_text(content),
            f"--{boundary}--",
            "",
        ]

        return (f"multipart/form-data; boundary={boundary}", "\r\n".join(lines))

    def request(self, url, data=None, method=None, content_type="application/json", additional_headers=None):
        if data and content_type == "application/json":
            data = json.dumps(data)

        headers = {}
        if isinstance(additional_headers, dict):
            headers = additional_headers.copy()

        # NOTE: fetch_url uses a password manager, which follows the
        # standard request-then-challenge basic-auth semantics. However as
        # JIRA allows some unauthorised operations it doesn't necessarily
        # send the challenge, so the request occurs as the anonymous user,
        # resulting in unexpected results. To work around this we manually
        # inject the auth header up-front to ensure that JIRA treats
        # the requests as authorized for this user.

        if self.vars.token is not None:
            headers.update(
                {
                    "Content-Type": content_type,
                    "Authorization": f"Bearer {self.vars.token}",
                }
            )
        else:
            auth = to_text(
                base64.b64encode(to_bytes(f"{self.vars.username}:{self.vars.password}", errors="surrogate_or_strict"))
            )
            headers.update(
                {
                    "Content-Type": content_type,
                    "Authorization": f"Basic {auth}",
                }
            )

        response, info = fetch_url(
            self.module, url, data=data, method=method, timeout=self.vars.timeout, headers=headers
        )

        if info["status"] not in (200, 201, 204):
            error = None
            try:
                error = json.loads(info["body"])
            except Exception:
                msg = f'The request "{method} {url}" returned the unexpected status code {info["status"]} {info["msg"]}\n{info.get("body")}'
                self.module.fail_json(msg=msg, exception=traceback.format_exc())
            if error:
                msg = []
                for key in ("errorMessages", "errors"):
                    if error.get(key):
                        msg.append(str(error[key]))
                if msg:
                    self.module.fail_json(msg=", ".join(msg))
                self.module.fail_json(msg=f"{error}")
            # Fallback print body, if it can't be decoded
            self.module.fail_json(msg=to_native(info["body"]))

        body = response.read()

        if body:
            return json.loads(to_text(body, errors="surrogate_or_strict"))
        return {}

    def post(self, url, data, content_type="application/json", additional_headers=None):
        return self.request(
            url, data=data, method="POST", content_type=content_type, additional_headers=additional_headers
        )

    def put(self, url, data):
        return self.request(url, data=data, method="PUT")

    def get(self, url):
        return self.request(url)


def main():
    jira = JIRA()
    jira.run()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Akini Ross (@akinross) <akinross@cisco.com>
# Copyright: (c) 2025, Samita Bhattacharjee (@samiib) <samitab@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: ndo_template_deploy
version_added: "2.13.0"
short_description: Deploy templates to sites on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Deploy templates to sites.
- Prior to deploy or redeploy a schema validation is executed.
- When schema validation fails, M(cisco.mso.ndo_template_deploy) fails and deploy or redeploy will not be executed.
- Only supports NDO v3.7 and higher
author:
- Akini Ross (@akinross)
- Samita Bhattacharjee (@samiib)
options:
  schema:
    description:
    - The name of the schema.
    type: str
  schema_id:
    description:
    - The ID of the schema.
    type: str
  template:
    description:
    - The name of the template.
    - This parameter or O(template_id) is required.
    type: str
  template_id:
    description:
    - The ID of the template.
    - This parameter or O(template) is required.
    type: str
  template_type:
    description:
    - The type of the template.
    - O(template_type) is ignored when O(template_id) is provided.
    type: str
    aliases: [ type ]
    default: application
    choices: [ tenant, l3out, application, fabric_policy, fabric_resource, monitoring_tenant, monitoring_access, service_device ]
  sites:
    description:
    - The list of site names where the template will be undeployed.
    - This option is required when O(state=undeploy).
    type: list
    elements: str
  state:
    description:
    - Use C(deploy) to deploy a template.
    - Use C(redeploy) to redeploy a template.
    - Use C(undeploy) to undeploy a template from specified O(sites).
    - Use C(undeploy_all) to undeploy a template from all sites.
    - Use C(query) to get deployment status.
    type: str
    choices: [ deploy, redeploy, undeploy, undeploy_all, query ]
    default: deploy
notes:
- The O(template) must exist before using this module in your playbook.
  Use M(cisco.mso.ndo_template) to create non-schema templates.
  Use M(cisco.mso.mso_schema_template) to create schema/application templates.
seealso:
- module: cisco.mso.mso_schema_site
- module: cisco.mso.mso_schema_template
- module: cisco.mso.ndo_template
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Deploy a schema template
  cisco.mso.ndo_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    state: deploy

- name: Deploy a fabric policy template
  cisco.mso.ndo_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: Fabric Policy 1
    type: fabric_policy
    state: deploy

- name: Redeploy a schema template
  cisco.mso.ndo_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    state: redeploy

- name: Redeploy a fabric policy template using template id
  cisco.mso.ndo_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template_id: '{{ fabric_policy_1.current.templateId }}'
    state: deploy

- name: Undeploy a schema template
  cisco.mso.ndo_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    sites: [Site1, Site2]
    state: undeploy

- name: Undeploy a fabric policy template from all sites
  cisco.mso.ndo_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: Fabric Policy 1
    type: fabric_policy
    state: undeploy_all

- name: Undeploy a fabric policy template from one site
  cisco.mso.ndo_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: Fabric Policy 1
    type: fabric_policy
    sites: [Site1]
    state: undeploy

- name: Query a schema template deploy status
  cisco.mso.ndo_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    state: query
  register: query_result

- name: Query a fabric policy template deploy status
  cisco.mso.ndo_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: Fabric Policy 1
    type: fabric_policy
    state: query
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import (
    MSOModule,
    mso_argument_spec,
)
from ansible_collections.cisco.mso.plugins.module_utils.template import MSOTemplate
from ansible_collections.cisco.mso.plugins.module_utils.constants import TEMPLATE_TYPES


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str"),
        schema_id=dict(type="str"),
        template=dict(type="str"),
        template_id=dict(type="str"),
        template_type=dict(type="str", choices=list(TEMPLATE_TYPES), default="application", aliases=["type"]),
        sites=dict(type="list", elements="str"),
        state=dict(
            type="str",
            default="deploy",
            choices=["deploy", "redeploy", "undeploy", "undeploy_all", "query"],
        ),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "undeploy", ["sites"]],
        ],
        mutually_exclusive=[
            ("template", "template_id"),
            ("schema", "schema_id"),
        ],
        required_one_of=[["template", "template_id"]],
    )

    schema = module.params.get("schema")
    schema_id = module.params.get("schema_id")
    template = module.params.get("template")
    template_id = module.params.get("template_id")
    template_type = module.params.get("template_type")
    sites = module.params.get("sites")
    state = module.params.get("state")

    mso = MSOModule(module)

    if template_id is not None:
        template_type = None
    elif template_type == "application":
        template = template.replace(" ", "")

    mso_template = MSOTemplate(mso, template_type, template, template_id, schema, schema_id)

    if template is None:
        template = mso_template.template_name
    if template_id is None:
        template_id = mso_template.template_id
    if template_type is None:
        template_type = mso_template.template_type
    else:
        template_type = TEMPLATE_TYPES[template_type]["template_type"]

    mso_template.validate_template(template_type)

    is_application = template_type == "application"
    if is_application:
        if schema_id is None and schema is None:
            schema_id = mso_template.schema_id
        elif schema is not None and schema_id is None:
            schema_id = mso.lookup_schema(schema)

    path = None
    if state == "query":
        if is_application:
            path = "status/schema/{0}/template/{1}".format(schema_id, template)
        elif mso_template.deploy_task_id:
            path = "task/{0}".format(mso_template.deploy_task_id)
        method = "GET"
        payload = None
    else:
        path = "task"
        method = "POST"
        is_undeploy = state == "undeploy" or state == "undeploy_all"
        if is_application:
            payload = dict(schemaId=schema_id, templateName=template)
        else:
            payload = dict(templateId=template_id)
        if state == "deploy":
            payload.update(isRedeploy=False)
        elif state == "redeploy":
            payload.update(isRedeploy=True)
        elif is_undeploy:
            payload.update(undeploy=get_site_ids(mso, mso_template, sites, state))
        if is_application and not is_undeploy:
            mso.validate_schema(schema_id)
    if not module.check_mode:
        mso.existing = mso.request(path, method=method, data=payload) if path else {}
    mso.exit_json()


def get_site_ids(mso, mso_template, sites, state):
    if state == "undeploy_all":
        return mso_template.deployed_site_ids
    return [site.get("siteId") for site in mso.lookup_sites(sites)] if sites else []


if __name__ == "__main__":
    main()



# Copyright (c) 2022, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

from ansible.module_utils.parsing.convert_bool import boolean

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule


@cmd_runner_fmt.unpack_args
def _values_fmt(values, value_types):
    result = []
    for value, value_type in zip(values, value_types):
        if value_type == "bool":
            value = "true" if boolean(value) else "false"
        result.extend(["--type", f"{value_type}", "--set", f"{value}"])
    return result


def xfconf_runner(module: AnsibleModule, **kwargs) -> CmdRunner:
    runner = CmdRunner(
        module,
        command="xfconf-query",
        arg_formats=dict(
            channel=cmd_runner_fmt.as_opt_val("--channel"),
            property=cmd_runner_fmt.as_opt_val("--property"),
            force_array=cmd_runner_fmt.as_bool("--force-array"),
            reset=cmd_runner_fmt.as_bool("--reset"),
            create=cmd_runner_fmt.as_bool("--create"),
            list_arg=cmd_runner_fmt.as_bool("--list"),
            values_and_types=_values_fmt,
            version=cmd_runner_fmt.as_fixed("--version"),
        ),
        **kwargs,
    )
    return runner


def get_xfconf_version(runner: CmdRunner) -> str:
    with runner("version") as ctx:
        rc, out, err = ctx.run()
        return out.splitlines()[0].split()[1]



#!/usr/bin/python

# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_clientscope

short_description: Allows administration of Keycloak client_scopes using Keycloak API

version_added: 3.4.0

description:
  - This module allows you to add, remove or modify Keycloak client_scopes using the Keycloak REST API. It requires access
    to the REST API using OpenID Connect; the user connecting and the client being used must have the requisite access rights.
    In a default Keycloak installation, admin-cli and an admin user would work, as would a separate client definition with
    the scope tailored to your needs and a user having the expected roles.
  - The names of module options are snake_cased versions of the camelCase ones found in the Keycloak API and its documentation
    at U(https://www.keycloak.org/docs-api/8.0/rest-api/index.html).
  - Attributes are multi-valued in the Keycloak API. All attributes are lists of individual values and are returned that way
    by this module. You may pass single values for attributes when calling the module, and this is translated into a list
    suitable for the API.
  - When updating a client_scope, where possible provide the client_scope ID to the module. This removes a lookup to the API
    to translate the name into the client_scope ID.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  action_group:
    version_added: 10.2.0

options:
  state:
    description:
      - State of the client_scope.
      - On V(present), the client_scope is created if it does not yet exist, or updated with the parameters you provide.
      - On V(absent), the client_scope is removed if it exists.
    default: 'present'
    type: str
    choices:
      - present
      - absent

  name:
    type: str
    description:
      - Name of the client_scope.
      - This parameter is required only when creating or updating the client_scope.
  realm:
    type: str
    description:
      - They Keycloak realm under which this client_scope resides.
    default: 'master'

  id:
    type: str
    description:
      - The unique identifier for this client_scope.
      - This parameter is not required for updating or deleting a client_scope but providing it reduces the number of API
        calls required.
  description:
    type: str
    description:
      - Description for this client_scope.
      - This parameter is not required for updating or deleting a client_scope.
  protocol:
    description:
      - Type of client.
      - The V(docker-v2) value was added in community.general 8.6.0.
    choices: ['openid-connect', 'saml', 'wsfed', 'docker-v2']
    type: str

  protocol_mappers:
    description:
      - A list of dicts defining protocol mappers for this client.
      - This is C(protocolMappers) in the Keycloak REST API.
    aliases:
      - protocolMappers
    type: list
    elements: dict
    suboptions:
      protocol:
        description:
          - This specifies for which protocol this protocol mapper.
          - Is active.
        choices: ['openid-connect', 'saml', 'wsfed', 'docker-v2']
        type: str

      protocolMapper:
        description:
          - 'The Keycloak-internal name of the type of this protocol-mapper. While an exhaustive list is impossible to provide
            since this may be extended through SPIs by the user of Keycloak, by default Keycloak as of 3.4 ships with at least:'
          - V(docker-v2-allow-all-mapper).
          - V(oidc-address-mapper).
          - V(oidc-full-name-mapper).
          - V(oidc-group-membership-mapper).
          - V(oidc-hardcoded-claim-mapper).
          - V(oidc-hardcoded-role-mapper).
          - V(oidc-role-name-mapper).
          - V(oidc-script-based-protocol-mapper).
          - V(oidc-sha256-pairwise-sub-mapper).
          - V(oidc-usermodel-attribute-mapper).
          - V(oidc-usermodel-client-role-mapper).
          - V(oidc-usermodel-property-mapper).
          - V(oidc-usermodel-realm-role-mapper).
          - V(oidc-usersessionmodel-note-mapper).
          - V(saml-group-membership-mapper).
          - V(saml-hardcode-attribute-mapper).
          - V(saml-hardcode-role-mapper).
          - V(saml-role-list-mapper).
          - V(saml-role-name-mapper).
          - V(saml-user-attribute-mapper).
          - V(saml-user-property-mapper).
          - V(saml-user-session-note-mapper).
          - An exhaustive list of available mappers on your installation can be obtained on the admin console by going to
            Server Info -> Providers and looking under 'protocol-mapper'.
        type: str

      name:
        description:
          - The name of this protocol mapper.
        type: str

      id:
        description:
          - Usually a UUID specifying the internal ID of this protocol mapper instance.
        type: str

      config:
        description:
          - Dict specifying the configuration options for the protocol mapper; the contents differ depending on the value
            of O(protocol_mappers[].protocolMapper) and are not documented other than by the source of the mappers and its
            parent class(es). An example is given below. It is easiest to obtain valid config values by dumping an already-existing
            protocol mapper configuration through check-mode in the RV(existing) return value.
        type: dict

  attributes:
    type: dict
    description:
      - A dict of key/value pairs to set as custom attributes for the client_scope.
      - Values may be single values (for example a string) or a list of strings.
  protocol_mappers_behavior:
    description:
      - Determine how O(protocol_mappers) behave when updating an existing client_scope.
    choices:
      subset:
        - Add missing protocol mappers, do not remove any missing mappers.
      exact:
        - Make the protocol mappers exactly as specified, adding and removing mappers as needed.
    aliases:
      - protocolMappersBehavior
    default: 'subset'
    type: str
    version_added: "13.1.0"
extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes

author:
  - Gaëtan Daubresse (@Gaetan2907)
"""

EXAMPLES = r"""
- name: Create a Keycloak client_scopes, authentication with credentials
  community.general.keycloak_clientscope:
    name: my-new-kc-clientscope
    realm: MyCustomRealm
    state: present
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
  delegate_to: localhost

- name: Create a Keycloak client_scopes, authentication with token
  community.general.keycloak_clientscope:
    name: my-new-kc-clientscope
    realm: MyCustomRealm
    state: present
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    token: TOKEN
  delegate_to: localhost

- name: Delete a keycloak client_scopes
  community.general.keycloak_clientscope:
    id: '9d59aa76-2755-48c6-b1af-beb70a82c3cd'
    state: absent
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
  delegate_to: localhost

- name: Delete a Keycloak client_scope based on name
  community.general.keycloak_clientscope:
    name: my-clientscope-for-deletion
    state: absent
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
  delegate_to: localhost

- name: Update the name of a Keycloak client_scope
  community.general.keycloak_clientscope:
    id: '9d59aa76-2755-48c6-b1af-beb70a82c3cd'
    name: an-updated-kc-clientscope-name
    state: present
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
  delegate_to: localhost

- name: Create a Keycloak client_scope with some custom attributes
  community.general.keycloak_clientscope:
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    name: my-new_clientscope
    description: description-of-clientscope
    protocol: openid-connect
    protocol_mappers:
      - config:
          access.token.claim: true
          claim.name: "family_name"
          id.token.claim: true
          jsonType.label: String
          user.attribute: lastName
          userinfo.token.claim: true
        name: family name
        protocol: openid-connect
        protocolMapper: oidc-usermodel-property-mapper
      - config:
          attribute.name: Role
          attribute.nameformat: Basic
          single: false
        name: role list
        protocol: saml
        protocolMapper: saml-role-list-mapper
    attributes:
      attrib1: value1
      attrib2: value2
      attrib3:
        - with
        - numerous
        - individual
        - list
        - items
  delegate_to: localhost
"""

RETURN = r"""
msg:
  description: Message as to what action was taken.
  returned: always
  type: str
  sample: "Client_scope testclientscope has been updated"

proposed:
  description: Representation of proposed client scope.
  returned: always
  type: dict
  sample: {"clientId": "test"}

existing:
  description: Representation of existing client scope (sample is truncated).
  returned: always
  type: dict
  sample:
    {
      "adminUrl": "http://www.example.com/admin_url",
      "attributes": {
        "request.object.signature.alg": "RS256"
      }
    }

end_state:
  description: Representation of client scope after module execution (sample is truncated).
  returned: on success
  type: dict
  sample:
    {
      "adminUrl": "http://www.example.com/admin_url",
      "attributes": {
        "request.object.signature.alg": "RS256"
      }
    }
"""

import copy

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    camel,
    get_token,
    is_struct_included,
    keycloak_argument_spec,
)


def normalise_cr(clientscoperep, remove_ids=False):
    """Re-sorts any properties where the order so that diff's is minimised, and adds default values where appropriate so that the
    the change detection is more effective.

    :param clientscoperep: the clientscoperep dict to be sanitized
    :param remove_ids: If set to true, then the unique ID's of objects is removed to make the diff and checks for changed
                       not alert when the ID's of objects are not usually known, (e.g. for protocol_mappers)
    :return: normalised clientscoperep dict
    """
    # Avoid the dict passed in to be modified
    clientscoperep = copy.deepcopy(clientscoperep)

    if "protocolMappers" in clientscoperep:
        clientscoperep["protocolMappers"] = sorted(
            clientscoperep["protocolMappers"], key=lambda x: (x.get("name"), x.get("protocol"), x.get("protocolMapper"))
        )
        for mapper in clientscoperep["protocolMappers"]:
            if remove_ids:
                mapper.pop("id", None)

            # Set to a default value.
            mapper["consentRequired"] = mapper.get("consentRequired", False)

    return clientscoperep


def sanitize_cr(clientscoperep):
    """Removes probably sensitive details from a clientscoperep representation.

    :param clientscoperep: the clientscoperep dict to be sanitized
    :return: sanitized clientrep dict
    """
    result = copy.deepcopy(clientscoperep)
    if "secret" in result:
        result["secret"] = "no_log"
    if "attributes" in result:
        if "saml.signing.private.key" in result["attributes"]:
            result["attributes"]["saml.signing.private.key"] = "no_log"
    return normalise_cr(result)


def main():
    """
    Module execution

    :return:
    """
    argument_spec = keycloak_argument_spec()

    protmapper_spec = dict(
        id=dict(type="str"),
        name=dict(type="str"),
        protocol=dict(type="str", choices=["openid-connect", "saml", "wsfed", "docker-v2"]),
        protocolMapper=dict(type="str"),
        config=dict(type="dict"),
    )

    meta_args = dict(
        state=dict(default="present", choices=["present", "absent"]),
        realm=dict(default="master"),
        id=dict(type="str"),
        name=dict(type="str"),
        description=dict(type="str"),
        protocol=dict(type="str", choices=["openid-connect", "saml", "wsfed", "docker-v2"]),
        attributes=dict(type="dict"),
        protocol_mappers=dict(type="list", elements="dict", options=protmapper_spec, aliases=["protocolMappers"]),
        protocol_mappers_behavior=dict(
            type="str", aliases=["protocolMappersBehavior"], choices=["subset", "exact"], default="subset"
        ),
    )

    argument_spec.update(meta_args)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=(
            [
                ["id", "name"],
                ["token", "auth_realm", "auth_username", "auth_password", "auth_client_id", "auth_client_secret"],
            ]
        ),
        required_together=([["auth_username", "auth_password"]]),
        required_by={"refresh_token": "auth_realm"},
    )

    result = dict(changed=False, msg="", diff={}, proposed={}, existing={}, end_state={})

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    realm = module.params["realm"]
    state = module.params["state"]
    cid = module.params["id"]
    name = module.params["name"]
    protocol_mappers = module.params["protocol_mappers"]
    protocol_mappers_behavior = module.params["protocol_mappers_behavior"]

    # Filter and map the parameters names that apply to the client scope
    clientscope_params = [
        x
        for x in module.params
        if x not in list(keycloak_argument_spec().keys()) + ["state", "realm", "protocol_mappers_behavior"]
        and module.params[x] is not None
    ]

    # See if it already exists in Keycloak
    if cid is None:
        before_clientscope = kc.get_clientscope_by_name(name, realm=realm)
    else:
        before_clientscope = kc.get_clientscope_by_clientscopeid(cid, realm=realm)

    if before_clientscope is None:
        before_clientscope = {}

    before_mappers = before_clientscope["protocolMappers"] if "protocolMappers" in before_clientscope else []

    # Build a proposed changeset from parameters given to this module
    changeset = {}

    for clientscope_param in clientscope_params:
        new_param_value = module.params[clientscope_param]

        # Unfortunately, the ansible argument spec checker introduces variables with null values when
        # they are not specified
        if clientscope_param == "protocol_mappers":
            new_param_value = [{k: v for k, v in x.items() if v is not None} for x in new_param_value]
        changeset[camel(clientscope_param)] = new_param_value

    # Prepare the desired values using the existing values (non-existence results in a dict that is save to use as a basis)
    desired_clientscope = copy.deepcopy(before_clientscope)
    desired_clientscope.update(changeset)
    desired_mappers_names = [x["name"] for x in protocol_mappers] if protocol_mappers else []

    if protocol_mappers_behavior == "subset":
        # add exsisting mappers to desired object
        for mapper in before_mappers:
            if mapper["name"] not in desired_mappers_names:
                if "protocolMappers" not in desired_clientscope:
                    desired_clientscope["protocolMappers"] = []
                desired_clientscope["protocolMappers"].append(mapper)

    # Cater for when it doesn't exist (an empty dict)
    if not before_clientscope:
        if state == "absent":
            # Do nothing and exit
            if module._diff:
                result["diff"] = dict(before="", after="")
            result["changed"] = False
            result["end_state"] = {}
            result["msg"] = "Clientscope does not exist; doing nothing."
            module.exit_json(**result)

        # Process a creation
        result["changed"] = True

        if name is None:
            module.fail_json(msg="name must be specified when creating a new clientscope")

        if module._diff:
            result["diff"] = dict(before="", after=sanitize_cr(desired_clientscope))

        if module.check_mode:
            module.exit_json(**result)

        # create it
        kc.create_clientscope(desired_clientscope, realm=realm)
        after_clientscope = kc.get_clientscope_by_name(name, realm)

        result["end_state"] = sanitize_cr(after_clientscope)

        result["msg"] = f"Clientscope {after_clientscope['name']} has been created with ID {after_clientscope['id']}"

    else:
        if state == "present":
            # Process an update

            # no changes
            # remove ids for compare, problematic if desired has no ids set (not required),
            # normalize for consentRequired in protocolMappers
            if normalise_cr(desired_clientscope, remove_ids=True) == normalise_cr(before_clientscope, remove_ids=True):
                result["changed"] = False
                result["end_state"] = sanitize_cr(desired_clientscope)
                result["msg"] = f"No changes required to clientscope {before_clientscope['name']}."
                module.exit_json(**result)

            # doing an update
            result["changed"] = True

            if module._diff:
                result["diff"] = dict(before=sanitize_cr(before_clientscope), after=sanitize_cr(desired_clientscope))

            if module.check_mode:
                # We can only compare the current clientscope with the proposed updates we have
                before_norm = normalise_cr(before_clientscope, remove_ids=True)
                desired_norm = normalise_cr(desired_clientscope, remove_ids=True)
                if module._diff:
                    result["diff"] = dict(before=sanitize_cr(before_norm), after=sanitize_cr(desired_norm))
                result["changed"] = not is_struct_included(desired_norm, before_norm)
                module.exit_json(**result)

            # do the update
            kc.update_clientscope(desired_clientscope, realm=realm)

            # do the protocolmappers update
            if protocol_mappers is not None:
                for protocol_mapper in protocol_mappers:
                    # update if protocolmapper exist
                    current_protocolmapper = kc.get_clientscope_protocolmapper_by_name(
                        desired_clientscope["id"], protocol_mapper["name"], realm=realm
                    )
                    if current_protocolmapper is not None:
                        protocol_mapper["id"] = current_protocolmapper["id"]
                        kc.update_clientscope_protocolmapper(desired_clientscope["id"], protocol_mapper, realm=realm)
                    # create otherwise
                    else:
                        kc.create_clientscope_protocolmapper(desired_clientscope["id"], protocol_mapper, realm=realm)

            if protocol_mappers_behavior == "exact":
                # check if we need to delete protocol mappers on the server
                for mapper in before_mappers:
                    if mapper["name"] not in desired_mappers_names:
                        # raise Exception(mapper)
                        kc.delete_clientscope_protocolmapper(desired_clientscope["id"], mapper["id"], realm=realm)

            after_clientscope = kc.get_clientscope_by_clientscopeid(desired_clientscope["id"], realm=realm)

            result["end_state"] = after_clientscope

            result["msg"] = f"Clientscope {after_clientscope['id']} has been updated"
            module.exit_json(**result)

        else:
            # Process a deletion (because state was not 'present')
            result["changed"] = True

            if module._diff:
                result["diff"] = dict(before=sanitize_cr(before_clientscope), after="")

            if module.check_mode:
                module.exit_json(**result)

            # delete it
            cid = before_clientscope["id"]
            kc.delete_clientscope(cid=cid, realm=realm)

            result["end_state"] = {}

            result["msg"] = f"Clientscope {before_clientscope['name']} has been deleted"

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2023, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule


_alias_state_map = dict(
    present="alias",
    absent="unalias",
    info="aliases",
)

_state_map = dict(
    present="install",
    absent="remove",
    enabled="enable",
    disabled="disable",
    refresh="refresh",
)


def snap_runner(module: AnsibleModule, **kwargs) -> CmdRunner:
    runner = CmdRunner(
        module,
        "snap",
        arg_formats=dict(
            state_alias=cmd_runner_fmt.as_map(_alias_state_map),  # snap_alias only
            name=cmd_runner_fmt.as_list(),
            alias=cmd_runner_fmt.as_list(),  # snap_alias only
            state=cmd_runner_fmt.as_map(_state_map),
            _list=cmd_runner_fmt.as_fixed("list"),
            _set=cmd_runner_fmt.as_fixed("set"),
            get=cmd_runner_fmt.as_fixed(["get", "-d"]),
            classic=cmd_runner_fmt.as_bool("--classic"),
            channel=cmd_runner_fmt.as_func(lambda v: [] if v == "stable" else ["--channel", f"{v}"]),
            options=cmd_runner_fmt.as_list(),
            info=cmd_runner_fmt.as_fixed("info"),
            revision=cmd_runner_fmt.as_opt_val("--revision"),
            hold=cmd_runner_fmt.as_bool("--hold"),
            unhold=cmd_runner_fmt.as_bool("--unhold"),
            dangerous=cmd_runner_fmt.as_bool("--dangerous"),
            devmode=cmd_runner_fmt.as_bool("--devmode"),
            version=cmd_runner_fmt.as_fixed("version"),
            _connect=cmd_runner_fmt.as_func(lambda v: ["connect", v]),
            _disconnect=cmd_runner_fmt.as_func(lambda v: ["disconnect", v]),
            _connections=cmd_runner_fmt.as_fixed("connections"),
            slot=cmd_runner_fmt.as_list(),
        ),
        check_rc=False,
        **kwargs,
    )
    return runner


def get_version(runner: CmdRunner) -> dict[str, list[str]]:
    with runner("version") as ctx:
        rc, out, err = ctx.run()
    return dict(x.split() for x in out.splitlines() if len(x.split()) == 2)



#
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

"""
The parse_xml filter plugin
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
name: parse_xml
author: Ganesh Nalawade (@ganeshrn)
version_added: "1.0.0"
short_description: The parse_xml filter plugin.
description:
  - This filter will load the spec file and pass the command output
    through it, returning JSON output.
  - The YAML spec file defines how to parse the CLI output.
  - This plugin is deprecated and will be removed in a future release after 2027-02-01, please Use ansible.utils.cli_parse instead.
notes:
  - To convert the XML output of a network device command into structured JSON output.
options:
  output:
    description:
    - This source xml on which parse_xml invokes.
    type: raw
    required: True
  tmpl:
    description:
    - The spec file should be valid formatted YAML.
      It defines how to parse the XML output and return JSON data.
    - For example C(xml_data | ansible.netcommon.parse_xml(template.yml)),
      in this case C(xml_data) represents xml data option.
    type: str
"""

EXAMPLES = r"""
# Using parse_xml

# example_output.xml

# <?xml version="1.0" encoding="UTF-8"?>
# <rpc-reply message-id="urn:uuid:0cadb4e8-5bba-47f4-986e-72906227007f">
# 	<data>
# 		<ntp>
# 			<nodes>
# 				<node>
# 					<node>0/0/CPU0</node>
# 					<associations>
# 						<is-ntp-enabled>true</is-ntp-enabled>
# 						<sys-leap>ntp-leap-no-warning</sys-leap>
# 						<peer-summary-info>
# 							<peer-info-common>
# 								<host-mode>ntp-mode-client</host-mode>
# 								<is-configured>true</is-configured>
# 								<address>10.1.1.1</address>
# 								<reachability>0</reachability>
# 							</peer-info-common>
# 							<time-since>-1</time-since>
# 						</peer-summary-info>
# 						<peer-summary-info>
# 							<peer-info-common>
# 								<host-mode>ntp-mode-client</host-mode>
# 								<is-configured>true</is-configured>
# 								<address>172.16.252.29</address>
# 								<reachability>255</reachability>
# 							</peer-info-common>
# 							<time-since>991</time-since>
# 						</peer-summary-info>
# 					</associations>
# 				</node>
# 			</nodes>
# 		</ntp>
# 	</data>
# </rpc-reply>

# parse_xml.yml

# ---
# vars:
#   ntp_peers:
#     address: "{{ item.address }}"
#     reachability: "{{ item.reachability}}"
# keys:
#   result:
#     value: "{{ ntp_peers }}"
#     top: data/ntp/nodes/node/associations
#     items:
#       address: peer-summary-info/peer-info-common/address
#       reachability: peer-summary-info/peer-info-common/reachability


- name: Facts setup
  ansible.builtin.set_fact:
    xml: "{{ lookup('file', 'example_output.xml') }}"

- name: Parse xml invocation
  ansible.builtin.debug:
    msg: "{{ xml | ansible.netcommon.parse_xml('parse_xml.yml') }}"


# Task Output
# -----------
#
# TASK [set xml Data]
# ok: [host] => changed=false
#   ansible_facts:
#     xml: |-
#       <?xml version="1.0" encoding="UTF-8"?>
#       <rpc-reply message-id="urn:uuid:0cadb4e8-5bba-47f4-986e-72906227007f">
#               <data>
#                       <ntp>
#                               <nodes>
#                                       <node>
#                                               <node>0/0/CPU0</node>
#                                               <associations>
#                                                       <is-ntp-enabled>true</is-ntp-enabled>
#                                                       <sys-leap>ntp-leap-no-warning</sys-leap>
#                                                       <peer-summary-info>
#                                                               <peer-info-common>
#                                                                       <host-mode>ntp-mode-client</host-mode>
#                                                                       <is-configured>true</is-configured>
#                                                                       <address>10.1.1.1</address>
#                                                                       <reachability>0</reachability>
#                                                               </peer-info-common>
#                                                               <time-since>-1</time-since>
#                                                       </peer-summary-info>
#                                                       <peer-summary-info>
#                                                               <peer-info-common>
#                                                                       <host-mode>ntp-mode-client</host-mode>
#                                                                       <is-configured>true</is-configured>
#                                                                       <address>172.16.252.29</address>
#                                                                       <reachability>255</reachability>
#                                                               </peer-info-common>
#                                                               <time-since>991</time-since>
#                                                       </peer-summary-info>
#                                               </associations>
#                                       </node>
#                               </nodes>
#                       </ntp>
#               </data>
#       </rpc-reply>

# TASK [Parse Data]
# ok: [host] => changed=false
#   ansible_facts:
#     output:
#       result:
#       - address:
#         - 10.1.1.1
#         - 172.16.252.29
#         reachability:
#         - '0'
#         - '255'
"""

from ansible.errors import AnsibleFilterError
from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
    AnsibleArgSpecValidator,
)

from ansible_collections.ansible.netcommon.plugins.plugin_utils.argspec_filter_utils import (
    convert_to_native,
)
from ansible_collections.ansible.netcommon.plugins.plugin_utils.parse_xml import parse_xml


try:
    from jinja2.filters import pass_environment
except ImportError:
    from jinja2.filters import environmentfilter as pass_environment

from ansible.utils.display import Display


@pass_environment
def _parse_xml(*args, **kwargs):
    """parse xml"""

    display = Display()
    display.warning(
        "The 'parse_xml' filter is deprecated and will be removed in a future release "
        "after 2027-02-01. Use 'ansible.utils.cli_parse' instead."
    )
    keys = ["output", "tmpl"]
    data = dict(zip(keys, args[1:]))
    data.update(kwargs)
    data = convert_to_native(data)
    aav = AnsibleArgSpecValidator(data=data, schema=DOCUMENTATION, name="parse_xml")
    valid, errors, updated_data = aav.validate()
    if not valid:
        raise AnsibleFilterError(errors)
    return parse_xml(**updated_data)


class FilterModule(object):
    """parse_xml"""

    def filters(self):
        """a mapping of filter names to functions"""
        return {"parse_xml": _parse_xml}



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Kevin Breit (@kbreit) <kevin.breit@kevinbreit.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    "status": ['deprecated'],
    'supported_by': 'community'
}

DOCUMENTATION = r'''
---
module: meraki_mx_malware
short_description: Manage Malware Protection in the Meraki cloud
description:
- Fully configure malware protection in a Meraki environment.
notes:
- Some of the options are likely only used for developers within Meraki.
deprecated:
  removed_in: '3.0.0'
  why: Updated modules released with increased functionality
  alternative: cisco.meraki.networks_appliance_security_malware
options:
    state:
      description:
      - Specifies whether object should be queried, created/modified, or removed.
      choices: [absent, present, query]
      default: query
      type: str
    net_name:
      description:
      - Name of network which configuration is applied to.
      aliases: [network]
      type: str
    net_id:
      description:
      - ID of network which configuration is applied to.
      type: str
    allowed_urls:
      description:
      - List of URLs to whitelist.
      type: list
      elements: dict
      suboptions:
        url:
          description:
          - URL string to allow.
          type: str
        comment:
          description:
          - Human readable information about URL.
          type: str
    allowed_files:
      description:
      - List of files to whitelist.
      type: list
      elements: dict
      suboptions:
        sha256:
          description:
          - 256-bit hash of file.
          type: str
          aliases: [ hash ]
        comment:
          description:
          - Human readable information about file.
          type: str
    mode:
      description:
      - Enabled or disabled state of malware protection.
      choices: [disabled, enabled]
      type: str

author:
- Kevin Breit (@kbreit)
extends_documentation_fragment: cisco.meraki.meraki
'''

EXAMPLES = r'''
- name: Enable malware protection
  meraki_malware:
    auth_key: abc123
    state: present
    org_name: YourOrg
    net_name: YourNet
    mode: enabled
  delegate_to: localhost

- name: Set whitelisted url
  meraki_malware:
    auth_key: abc123
    state: present
    org_name: YourOrg
    net_name: YourNet
    mode: enabled
    allowed_urls:
      - url: www.ansible.com
        comment: Ansible
      - url: www.google.com
        comment: Google
  delegate_to: localhost

- name: Set whitelisted file
  meraki_malware:
    auth_key: abc123
    state: present
    org_name: YourOrg
    net_name: YourNet
    mode: enabled
    allowed_files:
      - sha256: e82c5f7d75004727e1f3b94426b9a11c8bc4c312a9170ac9a73abace40aef503
        comment: random zip
  delegate_to: localhost

- name: Get malware settings
  meraki_malware:
    auth_key: abc123
    state: query
    org_name: YourNet
    net_name: YourOrg
  delegate_to: localhost
'''

RETURN = r'''
data:
    description: List of administrators.
    returned: success
    type: complex
    contains:
        mode:
            description: Mode to enable or disable malware scanning.
            returned: success
            type: str
            sample: enabled
        allowed_files:
            description: List of files which are whitelisted.
            returned: success
            type: complex
            contains:
                sha256:
                    description: sha256 hash of whitelisted file.
                    returned: success
                    type: str
                    sample: e82c5f7d75004727e1f3b94426b9a11c8bc4c312a9170ac9a73abace40aef503
                comment:
                    description: Comment about the whitelisted entity
                    returned: success
                    type: str
                    sample: TPS report
        allowed_urls:
            description: List of URLs which are whitelisted.
            returned: success
            type: complex
            contains:
                url:
                    description: URL of whitelisted site.
                    returned: success
                    type: str
                    sample: site.com
                comment:
                    description: Comment about the whitelisted entity
                    returned: success
                    type: str
                    sample: Corporate HQ
'''

from ansible.module_utils.basic import AnsibleModule, json
from ansible_collections.cisco.meraki.plugins.module_utils.network.meraki.meraki import MerakiModule, meraki_argument_spec


def main():
    # define the available arguments/parameters that a user can pass to
    # the module

    urls_arg_spec = dict(url=dict(type='str'),
                         comment=dict(type='str'),
                         )

    files_arg_spec = dict(sha256=dict(type='str', aliases=['hash']),
                          comment=dict(type='str'),
                          )

    argument_spec = meraki_argument_spec()
    argument_spec.update(state=dict(type='str', choices=['absent', 'present', 'query'], default='query'),
                         net_name=dict(type='str', aliases=['network']),
                         net_id=dict(type='str'),
                         mode=dict(type='str', choices=['enabled', 'disabled']),
                         allowed_urls=dict(type='list', default=None, elements='dict', options=urls_arg_spec),
                         allowed_files=dict(type='list', default=None, elements='dict', options=files_arg_spec),
                         )

    # the AnsibleModule object will be our abstraction working with Ansible
    # this includes instantiation, a couple of common attr would be the
    # args/params passed to the execution, as well as if the module
    # supports check mode
    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True,
                           )
    meraki = MerakiModule(module, function='malware')

    meraki.params['follow_redirects'] = 'all'

    query_url = {'malware': '/networks/{net_id}/appliance/security/malware'}
    update_url = {'malware': '/networks/{net_id}/appliance/security/malware'}

    meraki.url_catalog['get_one'].update(query_url)
    meraki.url_catalog['update'] = update_url

    org_id = meraki.params['org_id']
    if org_id is None:
        org_id = meraki.get_org_id(meraki.params['org_name'])
    net_id = meraki.params['net_id']
    if net_id is None:
        nets = meraki.get_nets(org_id=org_id)
        net_id = meraki.get_net_id(net_name=meraki.params['net_name'], data=nets)

    # Check for argument completeness
    if meraki.params['state'] == 'present':
        if meraki.params['allowed_files'] is not None or meraki.params['allowed_urls'] is not None:
            if meraki.params['mode'] is None:
                meraki.fail_json(msg="mode must be set when allowed_files or allowed_urls is set.")

    # Assemble payload
    if meraki.params['state'] == 'present':
        payload = dict()
        if meraki.params['mode'] is not None:
            payload['mode'] = meraki.params['mode']
        if meraki.params['allowed_urls'] is not None:
            payload['allowedUrls'] = meraki.params['allowed_urls']
        if meraki.params['allowed_files'] is not None:
            payload['allowedFiles'] = meraki.params['allowed_files']

    if meraki.params['state'] == 'query':
        path = meraki.construct_path('get_one', net_id=net_id)
        data = meraki.request(path, method='GET')
        if meraki.status == 200:
            meraki.result['data'] = data
    elif meraki.params['state'] == 'present':
        path = meraki.construct_path('get_one', net_id=net_id)
        original = meraki.request(path, method='GET')
        if meraki.is_update_required(original, payload):
            if meraki.module.check_mode is True:
                meraki.generate_diff(original, payload)
                original.update(payload)
                meraki.result['data'] = original
                meraki.result['changed'] = True
                meraki.exit_json(**meraki.result)
            path = meraki.construct_path('update', net_id=net_id)
            data = meraki.request(path, method='PUT', payload=json.dumps(payload))
            if meraki.status == 200:
                meraki.generate_diff(original, data)
                meraki.result['data'] = data
                meraki.result['changed'] = True
        else:
            meraki.result['data'] = original

    # in the event of a successful module execution, you will want to
    # simple AnsibleModule.exit_json(), passing the key/value results
    meraki.exit_json(**meraki.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_component_info

short_description: Retrieve component info in Keycloak

version_added: 8.2.0

description:
  - This module retrieve information on component from Keycloak.
attributes:
  action_group:
    version_added: 10.2.0

options:
  realm:
    description:
      - The name of the realm.
    required: true
    type: str
  name:
    description:
      - Name of the Component.
    type: str
  provider_type:
    description:
      - Provider type of components.
      - 'Examples: V(org.keycloak.storage.UserStorageProvider), V(org.keycloak.services.clientregistration.policy.ClientRegistrationPolicy),
        V(org.keycloak.keys.KeyProvider), V(org.keycloak.userprofile.UserProfileProvider), V(org.keycloak.storage.ldap.mappers.LDAPStorageMapper).'
    type: str
  parent_id:
    description:
      - Container ID of the components.
    type: str


extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes
  - community.general._attributes.info_module

author:
  - Andre Desrosiers (@desand01)
"""

EXAMPLES = r"""
- name: Retrive info of a UserStorageProvider named myldap
  community.general.keycloak_component_info:
    auth_keycloak_url: http://localhost:8080/auth
    auth_username: admin
    auth_password: password
    auth_realm: master
    realm: myrealm
    name: myldap
    provider_type: org.keycloak.storage.UserStorageProvider

- name: Retrive key info component
  community.general.keycloak_component_info:
    auth_keycloak_url: http://localhost:8080/auth
    auth_username: admin
    auth_password: password
    auth_realm: master
    realm: myrealm
    name: rsa-enc-generated
    provider_type: org.keycloak.keys.KeyProvider

- name: Retrive all component from realm master
  community.general.keycloak_component_info:
    auth_keycloak_url: http://localhost:8080/auth
    auth_username: admin
    auth_password: password
    auth_realm: master
    realm: myrealm

- name: Retrive all sub components of parent component filter by type
  community.general.keycloak_component_info:
    auth_keycloak_url: http://localhost:8080/auth
    auth_username: admin
    auth_password: password
    auth_realm: master
    realm: myrealm
    parent_id: "075ef2fa-19fc-4a6d-bf4c-249f57365fd2"
    provider_type: "org.keycloak.storage.ldap.mappers.LDAPStorageMapper"
"""

RETURN = r"""
components:
  description: JSON representation of components.
  returned: always
  type: list
  elements: dict
"""

from urllib.parse import quote

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    get_token,
    keycloak_argument_spec,
)


def main():
    """
    Module execution

    :return:
    """
    argument_spec = keycloak_argument_spec()

    meta_args = dict(
        name=dict(type="str"),
        realm=dict(type="str", required=True),
        parent_id=dict(type="str"),
        provider_type=dict(type="str"),
    )

    argument_spec.update(meta_args)

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    result = dict(changed=False, components=[])

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    realm = module.params.get("realm")
    parentId = module.params.get("parent_id")
    name = module.params.get("name")
    providerType = module.params.get("provider_type")

    objRealm = kc.get_realm_by_id(realm)
    if not objRealm:
        module.fail_json(msg=f"Failed to retrive realm '{realm}'")

    filters = []

    if parentId:
        filters.append(f"parent={quote(parentId, safe='')}")
    else:
        filters.append(f"parent={quote(objRealm['id'], safe='')}")

    if name:
        filters.append(f"name={quote(name, safe='')}")
    if providerType:
        filters.append(f"type={quote(providerType, safe='')}")

    result["components"] = kc.get_components(filter="&".join(filters), realm=realm)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2025 Ansible community
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:
    # Use together with the community.general.redfish module utils' REDFISH_COMMON_ARGUMENT_SPEC
    DOCUMENTATION = r"""
options:
  validate_certs:
    description:
      - If V(false), TLS/SSL certificates are not validated.
      - Set this to V(true) to enable certificate checking. Should be used together with O(ca_path).
    type: bool
    default: false
  ca_path:
    description:
      - PEM formatted file that contains a CA certificate to be used for validation.
      - Only used if O(validate_certs=true).
    type: path
  ciphers:
    required: false
    description:
      - TLS/SSL Ciphers to use for the request.
      - When a list is provided, all ciphers are joined in order with V(:).
      - See the L(OpenSSL Cipher List Format,https://www.openssl.org/docs/manmaster/man1/openssl-ciphers.html#CIPHER-LIST-FORMAT)
        for more details.
      - The available ciphers is dependent on the Python and OpenSSL/LibreSSL versions.
    type: list
    elements: str
"""



# Copyright (c) 2024, kurokobo <kurokobo@protonmail.com>
# Copyright (c) 2014, Michael DeHaan <michael.dehaan@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: timestamp
type: stdout
short_description: Adds simple timestamp for each header
version_added: 9.0.0
description:
  - This callback adds simple timestamp for each header.
author: kurokobo (@kurokobo)
options:
  timezone:
    description:
      - Timezone to use for the timestamp in IANA time zone format.
      - For example V(America/New_York), V(Asia/Tokyo)). Ignored on Python < 3.9.
    ini:
      - section: callback_timestamp
        key: timezone
    env:
      - name: ANSIBLE_CALLBACK_TIMESTAMP_TIMEZONE
    type: string
  format_string:
    description:
      - Format of the timestamp shown to user in 1989 C standard format.
      - Refer to L(the Python documentation,https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes)
        for the available format codes.
    ini:
      - section: callback_timestamp
        key: format_string
    env:
      - name: ANSIBLE_CALLBACK_TIMESTAMP_FORMAT_STRING
    default: "%H:%M:%S"
    type: string
seealso:
  - plugin: ansible.posix.profile_tasks
    plugin_type: callback
    description: >-
      You can use P(ansible.posix.profile_tasks#callback) callback plugin to time individual tasks and overall execution time
      with detailed timestamps.
extends_documentation_fragment:
  - ansible.builtin.default_callback
  - ansible.builtin.result_format_callback
"""


import sys
import types
from datetime import datetime

from ansible.module_utils.common.text.converters import to_text
from ansible.plugins.callback.default import CallbackModule as Default
from ansible.utils.display import get_text_width

# Store whether the zoneinfo module is available
_ZONEINFO_AVAILABLE = sys.version_info >= (3, 9)


def get_datetime_now(tz):
    """
    Returns the current timestamp with the specified timezone
    """
    return datetime.now(tz=tz)


def banner(self, msg, color=None, cows=True):
    """
    Prints a header-looking line with cowsay or stars with length depending on terminal width (3 minimum) with trailing timestamp

    Based on the banner method of Display class from ansible.utils.display

    https://github.com/ansible/ansible/blob/4403519afe89138042108e237aef317fd5f09c33/lib/ansible/utils/display.py#L511
    """
    timestamp = get_datetime_now(self.timestamp_tzinfo).strftime(self.timestamp_format_string)
    timestamp_len = get_text_width(timestamp) + 1  # +1 for leading space

    msg = to_text(msg)
    if self.b_cowsay and cows:
        try:
            self.banner_cowsay(f"{msg} @ {timestamp}")
            return
        except OSError:
            self.warning("somebody cleverly deleted cowsay or something during the PB run.  heh.")

    msg = msg.strip()
    try:
        star_len = self.columns - get_text_width(msg) - timestamp_len
    except OSError:
        star_len = self.columns - len(msg) - timestamp_len
    if star_len <= 3:
        star_len = 3
    stars = "*" * star_len
    self.display(f"\n{msg} {stars} {timestamp}", color=color)


class CallbackModule(Default):
    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "stdout"
    CALLBACK_NAME = "community.general.timestamp"

    def __init__(self):
        super().__init__()

        # Replace the banner method of the display object with the custom one
        self._display.banner = types.MethodType(banner, self._display)

    def set_options(self, task_keys=None, var_options=None, direct=None):
        super().set_options(task_keys=task_keys, var_options=var_options, direct=direct)

        # Store zoneinfo for specified timezone if available
        tzinfo = None
        if _ZONEINFO_AVAILABLE and self.get_option("timezone"):
            from zoneinfo import ZoneInfo

            tzinfo = ZoneInfo(self.get_option("timezone"))

        # Inject options into the display object
        self._display.timestamp_tzinfo = tzinfo
        self._display.timestamp_format_string = self.get_option("format_string")



# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)


"""
The index_of lookup plugin
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type


DOCUMENTATION = """
    name: index_of
    author: Bradley Thornton (@cidrblock)
    version_added: "1.0.0"
    short_description: Find the indices of items in a list matching some criteria
    description:
      - This plugin returns the indices of items matching some criteria in a list.
      - When working with a list of dictionaries, the key to evaluate can be specified.
      - B(index_of) is also available as a B(filter plugin) for convenience.
      - Using the parameters below- C(lookup('ansible.utils.index_of', data, test, value, key, fail_on_missing, wantlist)).
    options:
      data:
        description:
          - A list of items to enumerate and test against.
        type: list
        elements: raw
        required: True
      test:
        description:
          - The name of the test to run against the list, a valid jinja2 test or ansible test plugin.
          - Jinja2 includes the following tests U(http://jinja.palletsprojects.com/templates/#builtin-tests).
          - An overview of tests included in ansible U(https://docs.ansible.com/ansible/latest/user_guide/playbooks_tests.html).
        type: str
        required: True
      value:
        description:
          - The value used to test each list item against.
          - Not required for simple tests (e.g. C(true), C(false), C(even), C(odd))
          - May be a C(string), C(boolean), C(number), C(regular expression) C(dict) and so on, depending on the B(test) used.
        type: raw
      key:
        description:
          - When the data provided is a list of dictionaries, run the test against this dictionary key.
          - When using a I(key), the I(data) must only contain dictionaries.
          - See I(fail_on_missing) below to determine the behaviour when the I(key) is missing from a dictionary in the I(data).
        type: str
      fail_on_missing:
        description:
        - When provided a list of dictionaries, fail if the key is missing from one or more of the dictionaries.
        type: bool
      wantlist:
        description:
          - When only a single entry in the I(data) is matched, the index of that entry is returned as an integer.
          - If set to C(True), the return value will always be a list, even if only a single entry is matched.
          - This can also be accomplished using C(query) or C(q) instead of C(lookup).
          - U(https://docs.ansible.com/ansible/latest/plugins/lookup.html)
        type: bool

    notes:
"""

EXAMPLES = r"""

#### Simple examples

- ansible.builtin.set_fact:
    data:
      - 1
      - 2
      - 3

- name: Find the index of 2
  ansible.builtin.set_fact:
    indices: "{{ lookup('ansible.utils.index_of', data, 'eq', 2) }}"

# TASK [Find the index of 2] *************************************************
# ok: [nxos101] => changed=false
#   ansible_facts:
#     indices: '1'

- name: Find the index of 2, ensure list is returned
  ansible.builtin.set_fact:
    indices: "{{ lookup('ansible.utils.index_of', data, 'eq', 2, wantlist=True) }}"

# TASK [Find the index of 2, ensure list is returned] ************************
# ok: [nxos101] => changed=false
#   ansible_facts:
#     indices:
#     - 1

- name: Find the index of 3 using the long format
  ansible.builtin.set_fact:
    indices: "{{ lookup('ansible.utils.index_of', data=data, test='eq', value=value, wantlist=True) }}"
  vars:
    value: 3

# TASK [Find the index of 3 using the long format] ***************************
# ok: [nxos101] => changed=false
#   ansible_facts:
#     indices:
#     - 2

- name: Find numbers greater than 1, using loop
  debug:
    msg: "{{ data[item] }} is {{ test }} than {{ value }}"
  loop: "{{ lookup('ansible.utils.index_of', data, test, value) }}"
  vars:
    test: '>'
    value: 1

# TASK [Find numbers great than 1, using loop] *******************************
# ok: [sw01] => (item=1) =>
#   msg: 2 is > than 1
# ok: [sw01] => (item=2) =>
#   msg: 3 is > than 1

- name: Find numbers greater than 1, using with
  debug:
    msg: "{{ data[item] }} is {{ params.test }} than {{ params.value }}"
  with_ansible.utils.index_of: "{{ params }}"
  vars:
    params:
      data: "{{ data }}"
      test: '>'
      value: 1

# TASK [Find numbers greater than 1, using with] *****************************
# ok: [nxos101] => (item=1) =>
#   msg: 2 is > than 1
# ok: [nxos101] => (item=2) =>
#   msg: 3 is > than 1


#### Working with lists of dictionaries

- ansible.builtin.set_fact:
    data:
      - name: sw01.example.lan
        type: switch
      - name: rtr01.example.lan
        type: router
      - name: fw01.example.corp
        type: firewall
      - name: fw02.example.corp
        type: firewall

- name: Find the index of all firewalls using the type key
  ansible.builtin.set_fact:
    firewalls: "{{ lookup('ansible.utils.index_of', data, 'eq', 'firewall', 'type') }}"

# TASK [Find the index of all firewalls using the type key] ******************
# ok: [nxos101] => changed=false
#   ansible_facts:
#     firewalls:
#     - 2
#     - 3

- name: Find the index of all firewalls, use in a loop
  debug:
    msg: "The type of {{ device_type }} at index {{ item }} has name {{ data[item].name }}."
  loop: "{{ lookup('ansible.utils.index_of', data, 'eq', device_type, 'type') }}"
  vars:
    device_type: firewall

# TASK [Find the index of all firewalls, use in a loop, as a filter] *********
# ok: [nxos101] => (item=2) =>
#   msg: The type of firewall at index 2 has name fw01.example.corp.
# ok: [nxos101] => (item=3) =>
#   msg: The type of firewall at index 3 has name fw02.example.corp.

- name: Find the index of all devices with a .corp name
  debug:
    msg: "The device named {{ data[item].name }} is a {{ data[item].type }}"
  loop: "{{ lookup('ansible.utils.index_of', data, 'regex', expression, 'name') }}"
  vars:
    expression: '\.corp$'

# TASK [Find the index of all devices with a .corp name] *********************
# ok: [nxos101] => (item=2) =>
#   msg: The device named fw01.example.corp is a firewall
# ok: [nxos101] => (item=3) =>
#   msg: The device named fw02.example.corp is a firewall


#### Working with complex structures from resource modules

- name: Retrieve the current L3 interface configuration
  cisco.nxos.nxos_l3_interfaces:
    state: gathered
  register: current_l3

# TASK [Retrieve the current L3 interface configuration] *********************
# ok: [sw01] => changed=false
#   gathered:
#   - name: Ethernet1/1
#   - name: Ethernet1/2
#   <...>
#   - name: Ethernet1/128
#   - ipv4:
#     - address: 192.168.101.14/24
#     name: mgmt0

- name: Find the indices interfaces with a 192.168.101.xx ip address
  ansible.builtin.set_fact:
    found: "{{ found + entry }}"
  with_indexed_items: "{{ current_l3.gathered }}"
  vars:
    found: []
    ip: '192.168.101.'
    address: "{{ lookup('ansible.utils.index_of', item.1.ipv4 | d([]), 'search', ip, 'address', wantlist=True) }}"
    entry:
      - interface_idx: "{{ item.0 }}"
        address_idxs: "{{ address }}"
  when: address

# TASK [debug] ***************************************************************
# ok: [sw01] =>
#   found:
#   - address_idxs:
#     - 0
#     interface_idx: '128'

- name: Show all interfaces and their address
  debug:
    msg: "{{ interface.name }} has ip {{ address }}"
  loop: "{{ found|subelements('address_idxs') }}"
  vars:
    interface: "{{ current_l3.gathered[item.0.interface_idx|int] }}"
    address: "{{ interface.ipv4[item.1].address }}"

# TASK [Show all interfaces and their address] *******************************
# ok: [nxos101] => (item=[{'interface_idx': '128', 'address_idxs': [0]}, 0]) =>
#   msg: mgmt0 has ip 192.168.101.14/24


#### Working with deeply nested data

- ansible.builtin.set_fact:
    data:
      interfaces:
        interface:
          - config:
              description: configured by Ansible - 1
              enabled: true
              loopback-mode: false
              mtu: 1024
              name: loopback0000
              type: eth
            name: loopback0000
            subinterfaces:
              subinterface:
                - config:
                    description: subinterface configured by Ansible - 1
                    enabled: true
                    index: 5
                  index: 5
                - config:
                    description: subinterface configured by Ansible - 2
                    enabled: false
                    index: 2
                  index: 2
          - config:
              description: configured by Ansible - 2
              enabled: false
              loopback-mode: false
              mtu: 2048
              name: loopback1111
              type: virt
            name: loopback1111
            subinterfaces:
              subinterface:
                - config:
                    description: subinterface configured by Ansible - 3
                    enabled: true
                    index: 10
                  index: 10
                - config:
                    description: subinterface configured by Ansible - 4
                    enabled: false
                    index: 3
                  index: 3


- name: Find the description of loopback111, subinterface index 10
  debug:
    msg: |-
      {{ data.interfaces.interface[int_idx|int]
          .subinterfaces.subinterface[subint_idx|int]
            .config.description }}
  vars:
    # the values to search for
    int_name: loopback1111
    sub_index: 10
    # retrieve the index in each nested list
    int_idx: |
      {{ lookup('ansible.utils.index_of',
             data.interfaces.interface,
                 'eq', int_name, 'name') }}
    subint_idx: |
      {{ lookup('ansible.utils.index_of',
             data.interfaces.interface[int_idx|int].subinterfaces.subinterface,
                 'eq', sub_index, 'index') }}

# TASK [Find the description of loopback111, subinterface index 10] ************
# ok: [sw01] =>
#   msg: subinterface configured by Ansible - 3
"""

RETURN = """
  _raw:
    description:
      - One or more zero-based indicies of the matching list items.
      - See C(wantlist) if a list is always required.
"""

from ansible.errors import AnsibleLookupError
from ansible.plugins.lookup import LookupBase

from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
    AnsibleArgSpecValidator,
)
from ansible_collections.ansible.utils.plugins.plugin_utils.index_of import index_of


class LookupModule(LookupBase):
    def run(self, terms, variables, **kwargs):
        if isinstance(terms, list):
            keys = [
                "data",
                "test",
                "value",
                "key",
                "fail_on_missing",
                "wantlist",
            ]
            terms = dict(zip(keys, terms))
        terms.update(kwargs)

        aav = AnsibleArgSpecValidator(data=terms, schema=DOCUMENTATION, name="index_of")
        valid, errors, updated_data = aav.validate()
        if not valid:
            raise AnsibleLookupError(errors)
        updated_data["wantlist"] = True
        updated_data["tests"] = self._templar.environment.tests
        res = index_of(**updated_data)
        return res



# Copyright (c) 2025, Sean McAvoy (@smcavoy) <seanmcavoy@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = """
---
module: lxd_storage_pool_info
short_description: Retrieve information about LXD storage pools
version_added: 12.1.0
description:
  - Retrieve information about LXD storage pools.
  - This module returns details about all storage pools or a specific storage pool.
author: "Sean McAvoy (@smcavoy)"
extends_documentation_fragment:
  - community.general._attributes
  - community.general._attributes.info_module
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    description:
      - Name of a specific storage pool to retrieve information about.
      - If not specified, information about all storage pools are returned.
    type: str
  type:
    description:
      - Filter storage pools by driver/type (for example V(dir), V(zfs), V(btrfs), V(lvm), or V(ceph)).
      - If not specified, all storage pool types are returned.
      - Can be a single type or a list of types.
    type: list
    elements: str
    default: []
  project:
    description:
      - 'Project of the storage pool.
        See U(https://documentation.ubuntu.com/lxd/en/latest/projects/).'
    type: str
  url:
    description:
      - The Unix domain socket path or the https URL for the LXD server.
    default: unix:/var/lib/lxd/unix.socket
    type: str
  snap_url:
    description:
      - The Unix domain socket path when LXD is installed by snap package manager.
    default: unix:/var/snap/lxd/common/lxd/unix.socket
    type: str
  client_key:
    description:
      - The client certificate key file path.
      - If not specified, it defaults to C($HOME/.config/lxc/client.key).
    aliases: [key_file]
    type: path
  client_cert:
    description:
      - The client certificate file path.
      - If not specified, it defaults to C($HOME/.config/lxc/client.crt).
    aliases: [cert_file]
    type: path
  trust_password:
    description:
      - The client trusted password.
      - 'You need to set this password on the LXD server before
        running this module using the following command:
        C(lxc config set core.trust_password <some random password>)
        See U(https://www.stgraber.org/2016/04/18/lxd-api-direct-interaction/).'
      - If O(trust_password) is set, this module sends a request for
        authentication before sending any requests.
    type: str
"""

EXAMPLES = """
- name: Get information about all storage pools
  community.general.lxd_storage_pool_info:
  register: result

- name: Get information about a specific storage pool
  community.general.lxd_storage_pool_info:
    name: default
  register: result

- name: Get information about all ZFS storage pools
  community.general.lxd_storage_pool_info:
    type:
      - zfs
      - ceph
  register: result

- name: Get information about ZFS and BTRFS storage pools
  community.general.lxd_storage_pool_info:
    type:
      - zfs
      - btrfs
  register: result

- name: Get storage pool information via HTTPS connection
  community.general.lxd_storage_pool_info:
    url: https://127.0.0.1:8443
    trust_password: mypassword
    name: default
  register: result

- name: Get storage pool information for a specific project
  community.general.lxd_storage_pool_info:
    project: myproject
  register: result
"""

RETURN = """
storage_pools:
  description: List of LXD storage pools.
  returned: success
  type: list
  elements: dict
  sample: [
    {
      "name": "default",
      "driver": "dir",
      "used_by": ["/1.0/instances/container1"],
      "config": {
        "source": "/var/lib/lxd/storage-pools/default"
      },
      "description": "Default storage pool",
      "status": "Created",
      "locations": ["none"]
    }
  ]
  contains:
    name:
      description: The name of the storage pool.
      type: str
      returned: success
    driver:
      description: The storage pool driver/type.
      type: str
      returned: success
    used_by:
      description: List of resources using this storage pool.
      type: list
      elements: str
      returned: success
    config:
      description: Configuration of the storage pool.
      type: dict
      returned: success
    description:
      description: Description of the storage pool.
      type: str
      returned: success
    status:
      description: Current status of the storage pool.
      type: str
      returned: success
    locations:
      description: Cluster member locations for the pool.
      type: list
      elements: str
      returned: success
logs:
  description: The logs of requests and responses.
  returned: when ansible-playbook is invoked with -vvvv.
  type: list
  elements: dict
  sample: [
    {
      "type": "sent request",
      "request": {
        "method": "GET",
        "url": "/1.0/storage-pools",
        "json": null,
        "timeout": null
      },
      "response": {
        "json": {"type": "sync", "status": "Success"}
      }
    }
  ]
"""

import os
import typing as t
from urllib.parse import urlencode

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._lxd import (
    LXDClient,
    LXDClientException,
    default_cert_file,
    default_key_file,
)

# ANSIBLE_LXD_DEFAULT_URL is a default value of the lxd endpoint
ANSIBLE_LXD_DEFAULT_URL = "unix:/var/lib/lxd/unix.socket"
ANSIBLE_LXD_DEFAULT_SNAP_URL = "unix:/var/snap/lxd/common/lxd/unix.socket"

# API endpoints
LXD_API_VERSION = "1.0"
LXD_API_STORAGE_POOLS_ENDPOINT = f"/{LXD_API_VERSION}/storage-pools"


class LXDStoragePoolInfo:
    def __init__(self, module: AnsibleModule) -> None:
        """Gather information about LXD storage pools.

        :param module: Processed Ansible Module.
        :type module: AnsibleModule
        """
        self.module = module
        self.name = self.module.params["name"]
        # pool_type is guaranteed to be a list (may be empty) due to default=[]
        self.pool_type = self.module.params["type"]
        self.project = self.module.params["project"]

        self.key_file = self.module.params["client_key"]
        if self.key_file is None:
            self.key_file = default_key_file()
        self.cert_file = self.module.params["client_cert"]
        if self.cert_file is None:
            self.cert_file = default_cert_file()
        self.debug = self.module._verbosity >= 4

        snap_socket_path = self.module.params["snap_url"]
        if snap_socket_path.startswith("unix:"):
            snap_socket_path = snap_socket_path[5:]

        if self.module.params["url"] != ANSIBLE_LXD_DEFAULT_URL:
            self.url = self.module.params["url"]
        elif os.path.exists(snap_socket_path):
            self.url = self.module.params["snap_url"]
        else:
            self.url = self.module.params["url"]

        try:
            self.client = LXDClient(
                self.url,
                key_file=self.key_file,
                cert_file=self.cert_file,
                debug=self.debug,
            )
        except LXDClientException as e:
            self._fail_from_lxd_exception(e)

        self.trust_password = self.module.params["trust_password"]

    def _fail_from_lxd_exception(self, exception: LXDClientException) -> t.NoReturn:
        """Build failure parameters from LXDClientException and fail.

        :param exception: The LXDClientException instance
        :type exception: LXDClientException
        """
        fail_params = {}
        if self.client.debug and "logs" in exception.kwargs:
            fail_params["logs"] = exception.kwargs["logs"]
        self.module.fail_json(msg=exception.msg, changed=False, **fail_params)

    def _build_url(self, endpoint: str) -> str:
        """Build URL with project parameter if specified."""
        if self.project:
            return f"{endpoint}?{urlencode({'project': self.project})}"
        return endpoint

    def _get_storage_pool_list(self, recursion: int = 0) -> list:
        """Get list of all storage pools.

        :param recursion: API recursion level (0 for URLs only, 1 for full objects)
        :type recursion: int
        :return: List of pool names (recursion=0) or pool objects (recursion=1)
        :rtype: list
        """
        endpoint = LXD_API_STORAGE_POOLS_ENDPOINT
        if recursion > 0:
            endpoint = f"{endpoint}?recursion={recursion}"
        url = self._build_url(endpoint)
        resp_json = self.client.do("GET", url, ok_error_codes=[])

        if resp_json["type"] == "error":
            self.module.fail_json(
                msg=f"Failed to retrieve storage pools: {resp_json.get('error', 'Unknown error')}",
                error_code=resp_json.get("error_code"),
            )

        metadata = resp_json.get("metadata", [])

        if recursion == 0:
            # With recursion=0: list of pool URLs like ['/1.0/storage-pools/default']
            # Extract the pool names from URLs
            return [url.split("/")[-1] for url in metadata]
        else:
            # With recursion=1: list of pool objects with full metadata
            return metadata

    def _get_storage_pool_info(self, pool_name: str) -> dict:
        """Get detailed information about a specific storage pool."""
        url = self._build_url(f"{LXD_API_STORAGE_POOLS_ENDPOINT}/{pool_name}")
        resp_json = self.client.do("GET", url, ok_error_codes=[404])

        if resp_json["type"] == "error":
            if resp_json.get("error_code") == 404:
                self.module.fail_json(msg=f'Storage pool "{pool_name}" not found')
            else:
                self.module.fail_json(
                    msg=f'Failed to retrieve storage pool "{pool_name}": {resp_json.get("error", "Unknown error")}',
                    error_code=resp_json.get("error_code"),
                )

        return resp_json.get("metadata", {})

    def get_storage_pools(self) -> list[dict]:
        """Retrieve storage pool information based on module parameters."""
        storage_pools = []

        if self.name:
            # Get information about a specific storage pool
            pool_info = self._get_storage_pool_info(self.name)
            storage_pools.append(pool_info)
        else:
            # Get information about all storage pools using recursion
            # This retrieves all pool details in a single API call instead of one call per pool
            storage_pools = self._get_storage_pool_list(recursion=1)

        # Filter by type if specified
        if self.pool_type:
            storage_pools = [pool for pool in storage_pools if pool.get("driver") in self.pool_type]

        return storage_pools

    def run(self) -> None:
        """Run the main method."""
        try:
            if self.trust_password is not None:
                self.client.authenticate(self.trust_password)

            storage_pools = self.get_storage_pools()

            result_json = {"storage_pools": storage_pools}
            if self.client.debug:
                result_json["logs"] = self.client.logs
            self.module.exit_json(**result_json)
        except LXDClientException as e:
            self._fail_from_lxd_exception(e)


def main() -> None:
    """Ansible Main module."""
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="str"),
            type=dict(type="list", elements="str", default=[]),
            project=dict(type="str"),
            url=dict(type="str", default=ANSIBLE_LXD_DEFAULT_URL),
            snap_url=dict(type="str", default=ANSIBLE_LXD_DEFAULT_SNAP_URL),
            client_key=dict(type="path", aliases=["key_file"]),
            client_cert=dict(type="path", aliases=["cert_file"]),
            trust_password=dict(type="str", no_log=True),
        ),
        supports_check_mode=True,
    )

    lxd_info = LXDStoragePoolInfo(module=module)
    lxd_info.run()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_vsan_policy_info
short_description: Gather information about VSAN Policies and their VSANs in Cisco Intersight
description:
  - Retrieve information about VSAN (Virtual Storage Area Network) Policies and their associated VSANs from L(Cisco Intersight,https://intersight.com).
  - Query policies by organization, policy name, or filter VSANs by name patterns.
  - Returns structured data combining policy metadata with detailed VSAN configurations.
  - Supports filtering by organization, policy name, and VSAN name.
  - If no filters are provided, all VSAN Policies and their VSANs will be returned.
  - Returns structured data with both policy information and associated VSAN details.
extends_documentation_fragment: intersight
options:
  organization:
    description:
      - The name of the organization to filter VSAN Policies by.
      - Use 'default' for the default organization.
      - When specified, only policies from this organization will be returned.
    type: str
  name:
    description:
      - The exact name of the VSAN Policy to retrieve information from.
      - When specified, only the matching policy and its VSANs will be returned.
      - Can be combined with vsan_name filter to get specific VSANs from a specific policy.
    type: str
  vsan_name:
    description:
      - Filter VSANs by name within the policies.
      - Can be used with or without policy name filtering.
      - Supports exact matching (e.g., "vsan_uplink_100" will match only "vsan_uplink_100").
      - When combined with policy name, filters VSANs within that specific policy.
    type: str
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
# Basic Usage Examples
- name: Fetch all VSAN Policies and their VSANs from all organizations
  cisco.intersight.intersight_vsan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
  register: all_vsan_policies

# Organization-specific Examples
- name: Fetch all VSAN Policies from the default organization
  cisco.intersight.intersight_vsan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"
  register: default_org_policies

- name: Fetch all VSAN Policies from a custom organization
  cisco.intersight.intersight_vsan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "Engineering"
  register: engineering_policies

# Specific Policy Examples
- name: Fetch a specific VSAN Policy by name with all its VSANs
  cisco.intersight.intersight_vsan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "datacenter-vsan-policy"
  register: specific_policy

# VSAN Filtering Examples
- name: Find all policies containing a specific VSAN
  cisco.intersight.intersight_vsan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    vsan_name: "vsan_uplink_100"
  register: policies_with_vsan_100

- name: Find specific VSAN in a specific policy
  cisco.intersight.intersight_vsan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "datacenter-vsan-policy"
    vsan_name: "vsan_storage_200"
  register: specific_vsan_in_policy
'''

RETURN = r'''
api_response:
  description:
    - The API response containing policy and VSAN information.
    - Returns a dictionary when querying a single policy or no policies found.
    - Returns a list when multiple policies are found.
  returned: always
  type: dict
  sample:
    # Single policy response (when name parameter is used or only one policy found)
    Name: "datacenter-vsan-policy"
    ObjectType: "fabric.FcNetworkPolicy"
    Moid: "12345678901234567890abcd"
    Description: "VSAN policy for datacenter SAN infrastructure"
    EnableTrunking: false
    Organization:
      Name: "default"
      ObjectType: "organization.Organization"
    Tags:
      - Key: "Site"
        Value: "DataCenter-A"
      - Key: "Environment"
        Value: "Production"
    vsans:
      - Name: "vsan_uplink_100"
        ObjectType: "fabric.Vsan"
        Moid: "vsan12345678901234567890"
        VsanId: 100
        FcoeVlan: 100
        VsanScope: "Uplink"
        FcNetworkPolicy:
          Moid: "12345678901234567890abcd"
          ObjectType: "fabric.FcNetworkPolicy"
      - Name: "vsan_storage_200"
        ObjectType: "fabric.Vsan"
        Moid: "vsan09876543210987654321"
        VsanId: 200
        FcoeVlan: 200
        VsanScope: "Storage"
        FcNetworkPolicy:
          Moid: "12345678901234567890abcd"
          ObjectType: "fabric.FcNetworkPolicy"
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def get_vsans_for_policy(intersight, policy_moid, vsan_name_filter=None):
    """
    Get all VSANs associated with a specific policy.

    Args:
        intersight: IntersightModule instance
        policy_moid: MOID of the policy to get VSANs for
        vsan_name_filter: Optional filter for VSAN names

    Returns:
        List of VSAN objects associated with the policy
    """
    # Build filter for VSANs associated with this policy
    filter_conditions = [f"FcNetworkPolicy.Moid eq '{policy_moid}'"]

    # Add VSAN name filter if provided (exact match only)
    if vsan_name_filter:
        filter_conditions.append(f"Name eq '{vsan_name_filter}'")

    query_params = {
        '$filter': ' and '.join(filter_conditions)
    }

    # Reset api_response before the API call to avoid previous responses
    intersight.result['api_response'] = {}

    # Get VSANs for this policy
    intersight.get_resource(
        resource_path='/fabric/Vsans',
        query_params=query_params,
        return_list=True
    )

    # Capture the response immediately before it gets overwritten
    vsans_response = intersight.result.get('api_response', [])

    # Ensure we return a list
    if isinstance(vsans_response, list):
        return vsans_response
    elif isinstance(vsans_response, dict):
        return [vsans_response]
    else:
        return []


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        organization=dict(type='str'),
        name=dict(type='str'),
        vsan_name=dict(type='str')
    )
    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    intersight = IntersightModule(module)

    # Resource path used to fetch policy info
    resource_path = '/fabric/FcNetworkPolicies'

    # Get query parameters for policies
    query_params = intersight.set_query_params()

    # Reset api_response before the API call to avoid previous responses
    intersight.result['api_response'] = {}

    # Get VSAN policies
    intersight.get_resource(
        resource_path=resource_path,
        query_params=query_params,
        return_list=True
    )

    policies = intersight.result.get('api_response', [])
    vsan_name_filter = module.params.get('vsan_name')

    # Build structured response
    structured_results = []

    # Ensure policies is always a list for iteration, even if a single dict is returned
    if isinstance(policies, dict):
        policies = [policies]
    elif not isinstance(policies, list):
        policies = []

    for policy in policies:
        policy_moid = policy.get('Moid')
        if not policy_moid:
            continue

        # Get VSANs for this policy
        vsans = get_vsans_for_policy(intersight, policy_moid, vsan_name_filter)

        # Embed VSANs in the policy response
        if vsans:
            policy['vsans'] = vsans

        structured_results.append(policy)

    # Create final response structure
    final_api_response = None

    # Set final response based on number of policies found
    if len(structured_results) == 1:
        # Single policy - return as dict
        final_api_response = structured_results[0]
    elif len(structured_results) > 1:
        # Multiple policies - return as list
        final_api_response = structured_results
    else:
        # No policies found - return empty structure
        final_api_response = {}

    # Use intersight.result and update api_response directly
    intersight.result['api_response'] = final_api_response
    final_result = intersight.result

    module.exit_json(**final_result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2017 Jon Meran <jonathan.meran@sonos.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: batch_job_definition
version_added: 1.0.0
short_description: Manage AWS Batch Job Definitions
description:
  - This module allows the management of AWS Batch Job Definitions.
  - It is idempotent and supports "Check" mode.
  - Use module M(community.aws.batch_compute_environment) to manage the compute
    environment, M(community.aws.batch_job_queue) to manage job queues, M(community.aws.batch_job_definition) to manage job definitions.
  - Prior to release 5.0.0 this module was called C(community.aws.aws_batch_job_definition).
    The usage did not change.
author:
  - Jon Meran (@jonmer85)
options:
  job_definition_arn:
    description:
      - The ARN for the job definition.
    type: str
  job_definition_name:
    description:
      - The name for the job definition.
    required: true
    type: str
  state:
    description:
      - Describes the desired state.
    default: "present"
    choices: ["present", "absent"]
    type: str
  type:
    description:
      - The type of job definition.
    required: true
    type: str
  parameters:
    description:
      - Default parameter substitution placeholders to set in the job definition. Parameters are specified as a
        key-value pair mapping. Parameters in a SubmitJob request override any corresponding parameter defaults from
        the job definition.
    type: dict
  image:
    description:
      - >
        The image used to start a container. This string is passed directly to the Docker daemon. Images in the Docker
        Hub registry are available by default. Other repositories are specified with C(repository-url/image-name:tag).
        Up to 255 letters (uppercase and lowercase), numbers, hyphens, underscores, colons, periods, forward slashes,
        and number signs are allowed. This parameter maps to Image in the Create a container section of the Docker
        Remote API and the IMAGE parameter of docker run.
    required: true
    type: str
  vcpus:
    description:
      - The number of vCPUs reserved for the container. This parameter maps to CpuShares in the Create a container
        section of the Docker Remote API and the --cpu-shares option to docker run. Each vCPU is equivalent to
        1,024 CPU shares.
    required: true
    type: int
  memory:
    description:
      - The hard limit (in MiB) of memory to present to the container. If your container attempts to exceed the memory
        specified here, the container is killed. This parameter maps to Memory in the Create a container section of the
        Docker Remote API and the --memory option to docker run.
    required: true
    type: int
  command:
    description:
      - The command that is passed to the container. This parameter maps to Cmd in the Create a container section of
        the Docker Remote API and the COMMAND parameter to docker run. For more information,
        see U(https://docs.docker.com/engine/reference/builder/#cmd).
    type: list
    elements: str
    default: []
  job_role_arn:
    description:
      - The Amazon Resource Name (ARN) of the IAM role that the container can assume for AWS permissions.
    type: str
  volumes:
    description:
      - A list of data volumes used in a job.
    suboptions:
      host:
        description:
          - The contents of the host parameter determine whether your data volume persists on the host container
            instance and where it is stored. If the host parameter is empty, then the Docker daemon assigns a host
            path for your data volume, but the data is not guaranteed to persist after the containers associated with
            it stop running.
            This is a dictionary with one property, sourcePath - The path on the host container
            instance that is presented to the container. If this parameter is empty,then the Docker daemon has assigned
            a host path for you. If the host parameter contains a sourcePath file location, then the data volume
            persists at the specified location on the host container instance until you delete it manually. If the
            sourcePath value does not exist on the host container instance, the Docker daemon creates it. If the
            location does exist, the contents of the source path folder are exported.
      name:
        description:
          - The name of the volume. Up to 255 letters (uppercase and lowercase), numbers, hyphens, and underscores are
            allowed. This name is referenced in the sourceVolume parameter of container definition mountPoints.
    type: list
    elements: dict
    default: []
  environment:
    description:
      - The environment variables to pass to a container. This parameter maps to Env in the Create a container section
        of the Docker Remote API and the --env option to docker run.
    suboptions:
      name:
        description:
          - The name of the key value pair. For environment variables, this is the name of the environment variable.
      value:
        description:
          - The value of the key value pair. For environment variables, this is the value of the environment variable.
    type: list
    elements: dict
    default: []
  mount_points:
    description:
      - The mount points for data volumes in your container. This parameter maps to Volumes in the Create a container
        section of the Docker Remote API and the --volume option to docker run.
    suboptions:
      containerPath:
        description:
          - The path on the container at which to mount the host volume.
      readOnly:
        description:
          - If this value is true , the container has read-only access to the volume; otherwise, the container can write
             to the volume. The default value is C(false).
      sourceVolume:
        description:
          - The name of the volume to mount.
    type: list
    elements: dict
    default: []
  readonly_root_filesystem:
    description:
      - When this parameter is true, the container is given read-only access to its root file system. This parameter
        maps to ReadonlyRootfs in the Create a container section of the Docker Remote API and the --read-only option
        to docker run.
    type: str
  privileged:
    description:
      - When this parameter is true, the container is given elevated privileges on the host container instance
        (similar to the root user). This parameter maps to Privileged in the Create a container section of the
        Docker Remote API and the --privileged option to docker run.
    type: str
  ulimits:
    description:
      - A list of ulimits to set in the container. This parameter maps to Ulimits in the Create a container section
        of the Docker Remote API and the --ulimit option to docker run.
    suboptions:
      hardLimit:
        description:
          - The hard limit for the ulimit type.
      name:
        description:
          - The type of the ulimit.
      softLimit:
        description:
          - The soft limit for the ulimit type.
    type: list
    elements: dict
    default: []
  user:
    description:
      - The user name to use inside the container. This parameter maps to User in the Create a container section of
        the Docker Remote API and the --user option to docker run.
    type: str
  attempts:
    description:
      - Retry strategy - The number of times to move a job to the RUNNABLE status. You may specify between 1 and 10
        attempts. If attempts is greater than one, the job is retried if it fails until it has moved to RUNNABLE that
        many times.
    type: int
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
---
- name: My Batch Job Definition
  community.aws.batch_job_definition:
    job_definition_name: My Batch Job Definition
    state: present
    type: container
    parameters:
      Param1: Val1
      Param2: Val2
    image: <Docker Image URL>
    vcpus: 1
    memory: 512
    command:
      - python
      - run_my_script.py
      - arg1
    job_role_arn: <Job Role ARN>
    attempts: 3
  register: job_definition_create_result

- name: show results
  ansible.builtin.debug: var=job_definition_create_result
"""

RETURN = r"""
---
output:
  description: "returns what action was taken, whether something was changed, invocation and response"
  returned: always
  sample:
    aws_batch_job_definition_action: none
    changed: false
    response:
      job_definition_arn: "arn:aws:batch:...."
      job_definition_name: <name>
      status: INACTIVE
      type: container
  type: dict
"""

try:
    from botocore.exceptions import BotoCoreError
    from botocore.exceptions import ClientError
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.batch import cc
from ansible_collections.amazon.aws.plugins.module_utils.batch import set_api_params

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule

# ---------------------------------------------------------------------------------------------------
#
#   Helper Functions & classes
#
# ---------------------------------------------------------------------------------------------------

# logger = logging.getLogger()
# logging.basicConfig(filename='ansible_debug.log')
# logger.setLevel(logging.DEBUG)


def validate_params(module, batch_client):
    """
    Performs basic parameter validation.

    :param module:
    :param batch_client:
    :return:
    """
    return


# ---------------------------------------------------------------------------------------------------
#
#   Batch Job Definition functions
#
# ---------------------------------------------------------------------------------------------------


def get_current_job_definition(module, batch_client):
    try:
        environments = batch_client.describe_job_definitions(jobDefinitionName=module.params["job_definition_name"])
        if len(environments["jobDefinitions"]) > 0:
            latest_revision = max(map(lambda d: d["revision"], environments["jobDefinitions"]))
            latest_definition = next(
                (x for x in environments["jobDefinitions"] if x["revision"] == latest_revision), None
            )
            return latest_definition
        return None
    except ClientError:
        return None


def create_job_definition(module, batch_client):
    """
    Adds a Batch job definition

    :param module:
    :param batch_client:
    :return:
    """

    changed = False

    # set API parameters
    api_params = set_api_params(module, get_base_params())
    container_properties_params = set_api_params(module, get_container_property_params())
    retry_strategy_params = set_api_params(module, get_retry_strategy_params())

    api_params["retryStrategy"] = retry_strategy_params
    api_params["containerProperties"] = container_properties_params

    try:
        if not module.check_mode:
            batch_client.register_job_definition(**api_params)
        changed = True
    except (BotoCoreError, ClientError) as e:
        module.fail_json_aws(e, msg="Error registering job definition")

    return changed


def get_retry_strategy_params():
    return ("attempts",)


def get_container_property_params():
    return (
        "image",
        "vcpus",
        "memory",
        "command",
        "job_role_arn",
        "volumes",
        "environment",
        "mount_points",
        "readonly_root_filesystem",
        "privileged",
        "ulimits",
        "user",
    )


def get_base_params():
    return "job_definition_name", "type", "parameters"


def get_compute_environment_order_list(module):
    compute_environment_order_list = []
    for ceo in module.params["compute_environment_order"]:
        compute_environment_order_list.append(dict(order=ceo["order"], computeEnvironment=ceo["compute_environment"]))
    return compute_environment_order_list


def remove_job_definition(module, batch_client):
    """
    Remove a Batch job definition

    :param module:
    :param batch_client:
    :return:
    """

    changed = False

    try:
        if not module.check_mode:
            batch_client.deregister_job_definition(jobDefinition=module.params["job_definition_arn"])
        changed = True
    except (BotoCoreError, ClientError) as e:
        module.fail_json_aws(e, msg="Error removing job definition")
    return changed


def job_definition_equal(module, current_definition):
    equal = True

    for param in get_base_params():
        if module.params.get(param) != current_definition.get(cc(param)):
            equal = False
            break

    for param in get_container_property_params():
        if module.params.get(param) != current_definition.get("containerProperties").get(cc(param)):
            equal = False
            break

    for param in get_retry_strategy_params():
        if module.params.get(param) != current_definition.get("retryStrategy").get(cc(param)):
            equal = False
            break

    return equal


def manage_state(module, batch_client):
    changed = False
    current_state = "absent"
    state = module.params["state"]
    action_taken = "none"
    response = None

    # check if the job definition exists
    current_job_definition = get_current_job_definition(module, batch_client)
    if current_job_definition:
        current_state = "present"

    if state == "present":
        if current_state == "present":
            # check if definition has changed and register a new version if necessary
            if not job_definition_equal(module, current_job_definition):
                create_job_definition(module, batch_client)
                action_taken = "updated with new version"
                changed = True
        else:
            # Create Job definition
            changed = create_job_definition(module, batch_client)
            action_taken = "added"

        response = get_current_job_definition(module, batch_client)
        if not response:
            module.fail_json(msg="Unable to get job definition information after creating/updating")
    else:
        if current_state == "present":
            # remove the Job definition
            changed = remove_job_definition(module, batch_client)
            action_taken = "deregistered"
    return dict(changed=changed, batch_job_definition_action=action_taken, response=response)


# ---------------------------------------------------------------------------------------------------
#
#   MAIN
#
# ---------------------------------------------------------------------------------------------------


def main():
    """
    Main entry point.

    :return dict: ansible facts
    """

    argument_spec = dict(
        state=dict(required=False, default="present", choices=["present", "absent"]),
        job_definition_name=dict(required=True),
        job_definition_arn=dict(),
        type=dict(required=True),
        parameters=dict(type="dict"),
        image=dict(required=True),
        vcpus=dict(type="int", required=True),
        memory=dict(type="int", required=True),
        command=dict(type="list", default=[], elements="str"),
        job_role_arn=dict(),
        volumes=dict(type="list", default=[], elements="dict"),
        environment=dict(type="list", default=[], elements="dict"),
        mount_points=dict(type="list", default=[], elements="dict"),
        readonly_root_filesystem=dict(),
        privileged=dict(),
        ulimits=dict(type="list", default=[], elements="dict"),
        user=dict(),
        attempts=dict(type="int"),
    )

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True)

    batch_client = module.client("batch")

    validate_params(module, batch_client)

    results = manage_state(module, batch_client)

    module.exit_json(**camel_dict_to_snake_dict(results))


if __name__ == "__main__":
    main()
